#
#     This file is part of CasADi.
#
#     CasADi -- A symbolic framework for dynamic optimization.
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl,
#                             KU Leuven. All rights reserved.
#     Copyright (C) 2011-2014 Greg Horn
#
#     CasADi is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     CasADi is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from casadi import *

import casadi as ca

import subprocess
import os
import unittest
from helpers import *

class Daebuildertests(casadiTestCase):

  def test_reference_fmus(self):
    if "ghc-filesystem" not in CasadiMeta.feature_list(): return
    for name in ["VanDerPol2","VanDerPol3"]:
        fmu_file = "../data/" + name + ".fmu"
        if not os.path.exists(fmu_file):
            print("Skipping test_reference_fmus, resource not available")
            return
        dae = DaeBuilder("car",fmu_file)
        dae.disp(True)
        x0 = SX.sym('x0')
        x1 = SX.sym('x1')
        mu = SX.sym("mu")
        
        x = vertcat(x0,x1)
        c = mu
        f = dae.create('f',['x'],['ode'])
        f_ref = Function('f',[x],[vertcat(x1,1 * ((1 - x0 * x0) * x1) - x0)])
        test_point = [vertcat(1.1,1.3)]
        self.checkfunction(f,f_ref,inputs=[vertcat(1.1,1.3)],digits=7)
        if not name.endswith("3"):
            self.check_serialize(f,inputs=test_point)
  
  def test_cstr(self):
    fmu_file = "../data/cstr.fmu"
    if not os.path.exists(fmu_file):
        print("Skipping test_fmu_zip, resource not available")
        return
    unzipped_name = "cstr"
    unzipped_path = os.path.join(os.getcwd(), unzipped_name)
    import shutil
    if os.path.isdir(unzipped_path): shutil.rmtree(unzipped_path)
    import zipfile
    with zipfile.ZipFile(fmu_file, 'r') as zip_ref:
        zip_ref.extractall(unzipped_name)
    dae = DaeBuilder("cstr",unzipped_name)
    f = dae.create('f',['x','u'],['ode'])
    with self.assertInException("No stats available"):
        f.stats()
    dae = None
    
    C_A = SX.sym('C_A')
    C_B = SX.sym('C_B')
    V = 1
    k = 0.5
    
    q_in = SX.sym('q_in')
    C_A_in = SX.sym('C_A_in')
    C_B_in = SX.sym('C_B_in')

    x = vertcat(C_A,C_B)
    u = vertcat(C_A_in,C_B_in,q_in)    
    ode = vertcat((q_in*(C_A_in - C_A) - V*k*C_A*C_B)/V,(q_in*(C_B_in - C_B) + V*k*C_A*C_B)/V)


    f_ref = Function('f',[x,u],[ode],['x','u'],['ode'])
    
    test_point = [vertcat(1.1,1.3),vertcat(1.7,1.11,1.13)]
    self.checkfunction(f,f_ref,inputs=test_point,digits=4,hessian=False,evals=1)
    print(f.stats())
          
  def test_rumoca(self):
    if "rumoca" not in CasadiMeta.feature_list(): return
    rumoca = os.path.join(GlobalOptions.getCasadiPath(),'rumoca')
    rumoca_exe = os.path.join(GlobalOptions.getCasadiPath(),'rumoca.exe')
    if os.path.exists(rumoca) or os.path.exists(rumoca_exe):
        p = subprocess.run([rumoca,"-t","../assets/casadi_daebuilder.jinja","-m", "../assets/hello_world.mo"]) 

  def test_fmu_zip(self):
    fmu_file = "../data/VanDerPol2.fmu"
    if not os.path.exists(fmu_file):
        print("Skipping test_fmu_zip, resource not available")
        return
    unzipped_name = "VanDerPol2"
    unzipped_path = os.path.join(os.getcwd(), unzipped_name)
    for serialize_mode in ["link","embed"]:
        for use_zip in [True]:#,False]:
            if use_zip:
                if "ghc-filesystem" in CasadiMeta.feature_list():
                    dae = DaeBuilder("cstr",fmu_file,{"resource_serialize_mode": serialize_mode})
                else:
                    with self.assertInException("passing fmu files to DaeBuilder is unsupported"):
                        dae = DaeBuilder("cstr",fmu_file)
                    continue
            else:
                import shutil
                if os.path.isdir(unzipped_path): shutil.rmtree(unzipped_path)
                import zipfile
                with zipfile.ZipFile(fmu_file, 'r') as zip_ref:
                    zip_ref.extractall(unzipped_name)
                print('Unzipped %s into %s' % (fmu_file, unzipped_path))
                dae = DaeBuilder("cstr",unzipped_path)
            dae.disp(True)
            f = dae.create('f',['x'],['ode'])
            dae = None
            
            x0 = SX.sym('x0')
            x1 = SX.sym('x1')
            mu = SX.sym("mu")
            
            x = vertcat(x0,x1)
            c = mu
            f_ref = Function('f',[x],[vertcat(x1,1 * ((1 - x0 * x0) * x1) - x0)])
            test_point = [vertcat(1.1,1.3)]

            self.checkfunction_light(f,f_ref,inputs=test_point,digits=7)
            
            f.save('f.casadi')
            
            f = None
            
            f = Function.load("f.casadi")
            self.checkfunction_light(f,f_ref,inputs=test_point,digits=7)
            
            # Type decay, so test twice
            f.save('f.casadi')
            self.checkfunction_light(f,f_ref,inputs=test_point,digits=7)
            f = None
            f = Function.load("f.casadi")
            self.checkfunction(f,f_ref,inputs=test_point,hessian=False,digits=7,evals=1)
  
  @requires_rootfinder("kinsol")
  @requires_rootfinder("newton")
  @requires_nlpsol("ipopt")
  @requires_integrator("cvodes")
  @memory_heavy()
  def test_fmu_demo(self):
        # Use FMU to create a CasADi/DaeBuilder instance

        from time import time

        t1 = time()
        fmu_file = '../data/vdp.fmu'
        if not os.path.exists(fmu_file):
            print("Skipping test_fmu_demo, resource not available")
            return
        dae = ca.DaeBuilder('vdp', fmu_file)
        dae.disp(True)

        print('FMU ({}) provides directional derivatives: {}'.format(fmu_file, dae.provides_directional_derivative()))

        # Get state vector, initial conditions, bounds
        x = dae.x()
        lbx = dae.min(x)
        ubx = dae.max(x)
        x0 = dae.start(x)
        print('x: ', x)
        print('lbx: ', lbx)
        print('ubx: ', ubx)
        print('x0: ', x0)

        # Get free control, bounds
        u = dae.u()
        lbu = dae.min(u)
        ubu = dae.max(u)
        print('u: ', u)
        print('lbu: ', lbu)
        print('ubu: ', ubu)

        # Evaluate ODE right-hand-side
        f = dae.create('f', ['x', 'u'], ['ode'])
        print(f)
        # Evaluate the function numerically
        u_test = 0.4
        xdot_test = f(x0, u_test)
        print('xdot_test: ', xdot_test)



        # Alternative ODE right-hand-side function with u fixed
        dae.set('u', 0.4)
        f_no_u = dae.create('f_no_u', ['x'], ['ode'])
        print(f_no_u)
        # Evaluate the function numerically
        xdot_test_no_u = f_no_u(x0)
        print('xdot_test_no_u: ', xdot_test_no_u)

        # Evaluate ODE right-hand-side with auxilliary field
        f_with_aux = dae.create('f_with_aux', ['x', 'u'], ['ode'],
                       dict(aux = ['x1', 'x2', 'u']))
        print(f_with_aux)
        # After evaluation, the aux variables are in the stats field:
        u_test = 0.4
        xdot_test_with_aux = f_with_aux(x0, u_test)
        print('xdot_test_with_aux: ', xdot_test_with_aux)
        print(f_with_aux.stats())

        # A function that calculates the ODE and two Jacobian blocks:
        J = dae.create('J', ['x', 'u'], ['ode', 'jac_ode_x', 'jac_ode_u'],
                      dict(verbose = True, parallelization = 'openmp'))
        print(J)

        # Evaluate Jacobian
        xdot_test, A_test, B_test = J(x0, u_test)
        print('xdot_test: ', xdot_test)
        print('A_test: ', A_test)
        print('B_test: ', B_test)

        # A function that calculates the ODE and two Jacobian blocks:
        J_fd = dae.create('J_fd', ['x', 'u'], ['ode', 'jac_ode_x', 'jac_ode_u'],
                      dict(verbose = True,
                           parallelization = 'openmp',
                           enable_ad = False))
        print(J_fd)

        # Evaluate Jacobian
        xdot_fd, A_fd, B_fd = J_fd(x0, u_test)
        print('xdot_fd: ', xdot_fd)
        print('A_fd: ', A_fd)
        print('B_fd: ', B_fd)


        # Create a function that calculate the Hessian of a linear
        # combination of "ode" with respect to x and u: 
        H = dae.create('H', ['x', 'u', 'adj_ode'],
                       ['jac_adj_x_x', 'jac_adj_x_u', 'jac_adj_u_u'],
                       dict(verbose = True, parallelization = 'openmp'))
        print(H)
        # Evaluate Hessian
        # Evaluate Jacobian
        H_xx, H_xu, H_uu = H(x0, u_test, 1)
        print('H_xx: ', H_xx)
        print('H_xu: ', H_xu)
        print('H_uu: ', H_uu)



        import numpy as np
        # Create an integrator for simulating the ODE over 10 s using CVODES
        T = 10.
        tgrid = np.linspace(0., T, 100)
        dae.set_min(x[0], -np.inf) # relax lower bound on x1
        daefun = dae.create('daefun')
        sim = ca.integrator('sim', 'cvodes', daefun, 0, tgrid)




        # Call integrator instance
        r = sim(x0 = x0, u = u_test)
        x_traj_sim = r['xf'].full()



        # Create a new Simulator with support for forward sensitivities
        fwd_sim = ca.integrator('fwd_sim', 'cvodes', daefun, 0, tgrid, dict(nfwd = 1, verbose = False, print_stats = True, second_order_correction = False))
        print(fwd_sim)
        # Let's calculate sensitivity w.r.t. x1, i.e. use a forward seed [1, 0] for x0:
        seed_x0 = ca.DM([1, 0])
        fwd_test = fwd_sim(x0 = ca.horzcat(x0, seed_x0),
                          u = ca.horzcat(u_test, 0))
        # The directional derivatives can be found in column 1, 3, 5, etc. of `xf` for the different grid points.
        # (column 0, 2, 4, etc corresponds to nondifferentiated xf, for each grid point)
        print('dxf/dx0 * seed_x0 [AD, forward] = ', fwd_test['xf'][:, 1::2])
        # We can compare this result with a finite difference perturbation
        pert = 1e-3
        pert_sim = sim(x0=x0 + pert * seed_x0, u=u_test)
        x_pert_sim = pert_sim['xf']
        print('dxf/dx0 * seed_x0 [FD] = ', (x_pert_sim - x_traj_sim) / pert)



        # Create a new Simulator with support for adjoint sensitivities
        adj_sim = ca.integrator('adj_sim', 'cvodes', daefun, 0, tgrid, dict(nadj = 1, verbose = False, print_stats = True, second_order_correction = True, print_time = True))
        print(adj_sim)
        # Let's calculate sensitivity of xf[0] w.r.t. u at all the grid points, i.e. use an adjoint seed [1, 0] for xf:
        seed_xf = ca.DM.zeros(2, 100)
        seed_xf[:, -1] =  ca.DM([1, 0])  # We seed the state at the last grid point only
        adj_test = adj_sim(x0 = x0, adj_xf = seed_xf, u = u_test)
        print('trans(dxf/du) * seed_xf (AD, adjoint) = ', adj_test['adj_u'])

        fwd_adj_sim = ca.integrator('fwd_adj_sim', 'cvodes', daefun, 0, tgrid, dict(nfwd = 1, nadj = 1, verbose = False, print_stats = True, second_order_correction = False))
        print(fwd_adj_sim)
        # Let's calculate sensitivity of d(xf[0])/dx1 w.r.t. u at all the grid points, i.e. combining the seeds for forward and adjoint sensitivitiy analysis above
        fwd_adj_test = fwd_adj_sim(x0 = ca.horzcat(x0, seed_x0), adj_xf = ca.horzcat(seed_xf, ca.DM.zeros(seed_xf.shape)), u = ca.horzcat(u_test, 0))
        print('d(trans(dxf/du) * seed_xf)/dx0 * seed_x0 (AD, forward-over-adjoint) = ', fwd_adj_test['adj_u'][:,1::2])

        # Number of integrator steps
        N = 50
        # Size of the finite elements
        h = T/N

        # Let us use a 4th order collocation discretization using Legendre roots, cf. 
        # Nonlinear Programming: Concepts, Algorithms, and Applications to Chemical Processes
        # by Lorenz Biegler (2010)
        d = 3
        # The roots can be queried from CasADi or looked up in the above textbook
        tau_root = np.append(0, ca.collocation_points(d, 'legendre'))
        # Coefficients of the collocation equation
        C = np.zeros((d+1,d+1))
        # Coefficients of the continuity equation
        D = np.zeros(d+1)
        # Coefficients of the quadrature function
        B = np.zeros(d+1)
        # Construct polynomial basis
        for j in range(d+1):
            # Construct Lagrange polynomials
            p = np.poly1d([1])
            for r in range(d+1):
                if r != j:
                    p *= np.poly1d([1, -tau_root[r]]) / (tau_root[j]-tau_root[r])
            # Evaluate at the final time to get the coefficients of the continuity equation
            D[j] = p(1.0)
            # Evaluate the time derivative to get the coefficients of the continuity equation
            pder = np.polyder(p)
            for r in range(d+1): C[j,r] = pder(tau_root[r])
            # Evaluate the integral of the polynomial to get the coefficients of the quadrature function
            pint = np.polyint(p)
            B[j] = pint(1.0)
        print('tau_root: ', tau_root)
        print('C: ', C)
        print('D: ', D)
        print('B: ', B)

        # Symbolic expression for the controls (piecewise constant)
        U = ca.MX.sym('U',len(u))
        # Symbolic expressions for the states at each collocation point
        X = [ca.MX.sym('X_' + str(j), len(x)) for j in range(d+1)]
        # Define the collocation equations
        g = []
        for j in range(1,d+1):
          # Expression for the state derivative at the collocation point
          xdot_j = 0
          for r in range (d+1): xdot_j += C[r,j]*X[r]
          # Append collocation equations
          g.append(h*f(X[j],U) - xdot_j)
        # Concatenate constraints
        g = ca.vertcat(*g)
        # Form a root-finding function, implicitly defining X[1:] as a function of U, X[0]
        X_unknown = ca.vertcat(*X[1:])
        rfp = ca.Function('rfp', [X_unknown, X[0], U], [g], ['V', 'X0', 'U'], ['res'])

        # We can solve this system of equations using Sundials/KINSOL
        ifcn = ca.rootfinder('ifcn', 'kinsol', rfp, dict(print_level = 1))
        # Take a single step of the integrator
        v0 = ca.repmat(x0, d, 1)
        v0 = ifcn(v0, x0, u_test)
        # State at each collocation point
        x_all = ca.reshape(v0, len(x), d)
        # Prepend initial state
        x_all = ca.horzcat(x0, x_all)
        # Also calculate the state at the end
        xf = ca.mtimes(x_all, D)
        # Print solution
        print('x_all: ', x_all)
        print('xf: ', xf)

        # We can also use CasADi's native Newton solver
        ifcn = ca.rootfinder('ifcn', 'newton', rfp, dict(print_iteration = True))
        # Take a single step of the integrator
        v0 = ca.repmat(x0, d, 1)
        v0 = ifcn(v0, x0, u_test)
        # State at each collocation point
        x_all = ca.reshape(v0, len(x), d)
        # Prepend initial state
        x_all = ca.horzcat(x0, x_all)
        # Also calculate the state at the end
        xf = ca.mtimes(x_all, D)
        # Print solution
        print('x_all: ', x_all)
        print('xf: ', xf)

        # We can solve this system of equations using Sundials/KINSOL
        print(rfp)
        ifcn = ca.rootfinder('ifcn', 'kinsol', rfp, dict(print_level = 1))
        print(ifcn)
        # Differentiate Newton solver
        jac_ifcn = ifcn.jacobian()
        sol = jac_ifcn(V0 = v0, X0 = x0, U = u_test)
        # Jacobian of state at collocation points w.r.t. U
        print('jac_V_U = ', sol['jac_V_U'])
        # Jacobian of state at collocation points w.r.t. X0
        print('jac_V_X0 = ', sol['jac_V_X0'])

        # Compare with finite differences
        pert = 1e-3
        # Perturn u
        v0_pert = ifcn(ca.repmat(x0, d, 1), x0, u_test + pert)
        jac_V_U_fd = (v0_pert - v0) / pert
        print('jac_V_U (FD) = ', jac_V_U_fd)
        # Perturb x1
        v0_pert = ifcn(ca.repmat(x0, d, 1), x0 + ca.DM([pert, 0]), u_test)
        jac_V_X0_fd1 = (v0_pert - v0) / pert
        # Perturb x2
        v0_pert = ifcn(ca.repmat(x0, d, 1), x0 + ca.DM([0, pert]), u_test)
        jac_V_X0_fd2 = (v0_pert - v0) / pert
        print('jac_V_X0 (FD) = ', ca.horzcat(jac_V_X0_fd1, jac_V_X0_fd2))

        # Let's create a CasADi function for simulating the whole trajectory using Kinsol
        ifcn = ca.rootfinder('ifcn', 'kinsol', rfp)
        x0_in = ca.MX.sym('x0', len(x))
        u_all = []
        xk = x0_in
        x_all = [x0_in]
        for k in range(N):
            # Symbolic expression for the control for the interval
            uk = ca.MX.sym('u' + str(k), len(u))
            u_all.append(uk)
            # Solve rootfinding problem, using previous step as guess
            vk = ifcn(ca.repmat(xk,d,1), xk, uk)
            # Reshape and prepend initial state
            vk = ca.reshape(vk, len(x), d)
            vk = ca.horzcat(xk, vk)
            # Get the state at the end
            xk = ca.mtimes(vk, D)
            # Save trajectory
            x_all.append(xk)
        # Embed in a CasADi Function
        irksim = ca.Function('irksim', [x0_in, ca.vcat(u_all)], [ca.hcat(x_all)],
                                 ['x0', 'u'], ['x'])
        # Evaluate the function
        x_traj = irksim(x0, u_test).full()

        # A nonlinear equation solver is a differentiable object in CasADi.
        # We can differentiate it analytically to get the Jacobian-times-vector product:
        fwd_irksim = irksim.forward(1)
        # Let's calculate sensitivity w.r.t. x1, i.e. use a forward seed [1, 0] for x0:
        seed_x0 = ca.DM([1, 0])
        fwd_test = fwd_irksim(x0 = x0, u = u_test, fwd_x0 = seed_x0)
        print('fwd_x (AD, forward) = ', fwd_test['fwd_x'])
        # We can compare this result with a finite difference perturbation
        pert = 1e-3
        x_pert = irksim(x0 + pert * seed_x0, u_test).full()
        print('fwd_x (FD) = ', (x_pert - x_traj) / pert)

        # Now let's to a reverse mode AD, i.e. a transposed-Jacobian-times-vector product
        adj_irksim = irksim.reverse(1)
        # We seed the last entry in x2
        seed_x = ca.DM.zeros(2, N + 1)
        seed_x[1, N] = 1
        adj_test = adj_irksim(x0 = x0, u = u_test, adj_x = seed_x)
        print('adj_x0 (AD, reverse) = ', adj_test['adj_x0'])
        print('adj_u (AD, reverse) = ', adj_test['adj_u'])

        # We can check that the result is consistent with forward mode AD:
        print('Forward mode: derivative of x[1, N] w.r.t. x0[0] = ', fwd_test['fwd_x'][1, N])
        print('Reverse mode: derivative of x[1, N] w.r.t. x0[0] = ', adj_test['adj_x0'][0])

        # We can also calculate the full Jacobian of x w.r.t. x0 and u:
        jac_irksim = irksim.jacobian()
        jac_test = jac_irksim(x0 = x0, u = u_test)
        print('jac_x_x0 = ', jac_test['jac_x_x0'])
        print('jac_x_u = ', jac_test['jac_x_u'])

        # We will construct an NLP incrementally, starting with no decision variables, 
        # and constraints, and zero objective
        w = []
        g = []
        J = 0
        # Loop over all control intervals, constructing the NLP and x, u trajectories
        U_all = []
        X_all = []
        xk = ca.MX(x0)
        for k in range(N):
            # Symbolic expression for the control for the interval
            uk = ca.MX.sym('u' + str(k), len(u))
            U_all.append(uk)
            w.append(uk)
            # States at each collocation point for the interval
            Xc = [xk]
            for j in range(1, d + 1):
                xkj = ca.MX.sym('x' + str(k) + '_' + str(j), len(x))
                Xc.append(xkj)
                X_all.append(xkj)
                w.append(xkj)
            # Collect the collocation equations
            for j in range(1,d+1):
              # Expression for the state derivative at the collocation point
              xdot_j = 0
              for r in range (d+1): xdot_j += C[r,j]*Xc[r]
              # Append collocation equations
              g.append(h*f(Xc[j],uk) - xdot_j)
            # State at the end of the interval
            xk_end = 0
            for j in range(0,d+1): xk_end = xk_end + D[j] * Xc[j]
            # New variable for the state at the end of the interval
            xk = ca.MX.sym('x' + str(k + 1), len(x))
            X_all.append(xk)
            w.append(xk)
            # Enforce continuity
            g.append(xk_end - xk)
            # Add contributions to the objective
            J = J + ca.sumsqr(uk) + ca.sumsqr(xk)
        # Concatenate vectors
        w = ca.vcat(w)
        g = ca.vcat(g)
        X_all = ca.hcat(X_all)
        U_all = ca.hcat(U_all)
        # Create mappings from (u, x) -> (w) and back
        to_w = ca.Function('to_w', [X_all, U_all], [w], ['x', 'u'], ['w'])
        from_w = ca.Function('from_w', [w], [X_all, U_all], ['w'], ['x', 'u'])
        # Form the NLP
        nlp = dict(x = w, f = J, g = g)


        # Let's form the Jacobian of g w.r.t. w
        t0 = time()
        jac_g_w = ca.jacobian(g, w)
        print('Symbolic expression for Jacobian of g w.r.t w formed in %g s' % (time() - t0))
        # Create a function for evaluation
        t0 = time()
        jfcn = ca.Function('jfcn', [w], [jac_g_w], ['w'], ['jac_g_w'])
        print('Jacobian function formed in %g s' % (time() - t0))
        # Evaluate function
        w_test = to_w(u = u_test, x = x0)['w']
        t0 = time()
        J_test = jfcn(w_test)
        print('Jacobian evaluated in %g s' % (time() - t0))

        # Create an IPOPT instance, using L-BFGS
        opts = dict(ipopt = dict(hessian_approximation = 'exact', linear_solver = 'mumps'))
        solver = ca.nlpsol('solver', 'ipopt', nlp, opts)



        # Generate initial guess for w
        w0 = to_w(u = 0, x = x0)['w']
        # Lower bound for w
        lbw = to_w(u = lbu, x = lbx)['w']
        # Upper bound for w
        ubw = to_w(u = ubu, x = ubx)['w']
        # Solve NLP
        sol = solver(x0 = w0, lbx = lbw, ubx = ubw, lbg = 0, ubg = 0)


if __name__ == '__main__':
    unittest.main()



# This is an automatically generated file converting from the apm format

import casadi as cs

def hs049():
    # The optimal objective is (if given in):
    x_opt = cs.DM([1,1,1,1,1])
    f_opt = 0
    x = cs.MX.sym('x', 5)
    obj = cs.MX.zeros(1, 1)
    x0 = cs.DM.zeros(5, 1)
    lbx = -cs.inf*cs.DM.ones(5, 1)
    ubx = cs.inf*cs.DM.ones(5, 1)
    g = cs.MX.zeros(2, 1)
    lbg = -cs.inf*cs.DM.ones(2, 1)
    ubg = cs.inf*cs.DM.ones(2, 1)
    
    x0[0]= 10
    x0[1]=  7
    x0[2]=  2
    x0[3]= -3
    x0[4]= 0.8
    
    lbg[0] =  7
    ubg[0] =  7
    g[0] = x[0]+x[1]+x[2]+4*x[3]
    lbg[1] =  6
    ubg[1] =  6
    g[1] = x[2] + 5*x[4]
    obj = (x[0]-x[1])**2 + (x[2]-1)**2 + (x[3]-1)**4 + (x[4]-1)**6
    
    
    f = obj
    
    return (x_opt, f_opt, x, f, g, lbg, ubg, lbx, ubx, x0)



#
#     This file is part of CasADi.
#
#     CasADi -- A symbolic framework for dynamic optimization.
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl,
#                             KU Leuven. All rights reserved.
#     Copyright (C) 2011-2014 Greg Horn
#
#     CasADi is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     CasADi is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from casadi import *
import casadi
import numpy
from numpy import random, array
import unittest
import sys
from math import isnan, isinf
import itertools
import time
from contextlib import contextmanager
from casadi.tools import capture_stdout
import os

codegen_check_digits = 15

print("os" in os.environ)
if "os" in os.environ:
    print(os.environ["os"])
    if os.environ["os"]=="osx_arm":
        # Numerical differences due to different compilers in ci
        codegen_check_digits = 10

print("codegen_check_digits", codegen_check_digits)

import argparse
import struct

if sys.version_info >= (3, 0):
  import builtins
else:
  import __builtin__
  builtins = __builtin__

x = MX.sym("x",2,3)
y = MX.sym("x",3,4)

try:
  x+y
except TypeError:
  swig4 = True
except RuntimeError:
  swig4 = False

try:
  nlpsol(123)
except Exception as e:
  if "SXDict" in str(e):
    systemswig = True
  elif "dict:SX" in str(e):
    systemswig = False
  else:
    systemswig = True

# Pending deprecation in numpy
check_matrix = False

platform_arch = 8 * struct.calcsize("P")

parser = argparse.ArgumentParser()
parser.add_argument('--known_bugs', help='Run with known bugs', action='store_true')
parser.add_argument('--ignore_memory_heavy', help='Skip those tests that have a high memory footprint', action='store_true')
parser.add_argument('--ignore_memory_light', help='Skip those tests that have a lightweight memory footprint', action='store_true')
parser.add_argument('--run_slow', help='Skip those tests that take a long time to run', action='store_true')
parser.add_argument('unittest_args', nargs='*')

args = parser.parse_args()

import sys
sys.argv[1:] = ['-v'] + args.unittest_args

from io import StringIO


class LazyString(object):
  def __init__(self,f):
     self.f = f
     self.context = dict(sys._getframe(1).f_locals)

  def __add__(self, other):
     return LazyString('str(self) + other')
  def __radd__(self, other):
     return LazyString('other + str(self)')

  def __str__(self):
    d = self.context
    exec("ret = " + self.f.replace("\n","\\n"), d)
    return str(d["ret"])

class TeeString(StringIO):
  def __init__(self,stream):
    StringIO.__init__(self)
    self.stream = stream

  def write(self, data):
    StringIO.write(self,data)
    self.stream.write(data)

class Stdout():
    def __init__(self):
        self.stream = TeeString(sys.stdout)

    def __enter__(self):
        sys.stdout = self.stream
        return self.stream

    def __exit__(self, type, value, traceback):
        sys.stdout = self.stream.stream

class Stderr():
    def __init__(self):
        self.stream = TeeString(sys.stderr)

    def __enter__(self):
        sys.stderr = self.stream
        return self.stream

    def __exit__(self, type, value, traceback):
        sys.stderr = self.stream.stream

class FunctionPool:
  def __init__(self):
    self.numpyoperators=[]
    self.casadioperators=[]
    self.names=[]
    self.flags=[]
  def append(self,cas,num,name="",flags=set()):
    self.casadioperators.append(cas)
    self.numpyoperators.append(num)
    self.names.append(name)
    self.flags.append(flags)
  def zip(self):
    return list(zip(self.casadioperators,self.numpyoperators,self.names,self.flags))


def toSX_fun(fun):
  ins = fun.sx_in()
  return Function("f",ins,fun.call(ins))

def toMX_fun(fun):
  ins = fun.mx_in()
  return Function("f",ins,fun(ins))

def jacobian_old(f, i, j):
    return f.factory(f.name() + '_jac', f.name_in(),
        ['jac:' + f.name_out(j) + ':' + f.name_in(i)] + f.name_out())

def hessian_old(f, i, j):
    return f.factory(f.name() + '_hess', f.name_in(),
        ['hess:' + f.name_out(j) + ':' + f.name_in(i) + ":" + f.name_in(i),
        'grad:' + f.name_out(j) + ':' + f.name_in(i)] + f.name_out())

class casadiTestCase(unittest.TestCase):

  @classmethod
  def tearDownClass(cls):
    print("STATUS_RAN_ALL_TESTS")

  @contextmanager
  def assertInAnyOutput(self,s):
    e = ""
    with capture_stdout() as result:
        try:
          yield
        except Exception as err:
          e = str(err)
    print(result[0])
    self.assertTrue(s in e or s in result[0] or s in result[1])

  @contextmanager
  def assertInException(self,s):
    e = None
    try:
      yield
    except Exception as err:
      e = str(err)
    self.assertFalse(e is None)
    self.assertTrue(s in e,msg=e + "<->" + s)

  @contextmanager
  def assertOutput(self,included,excluded):
    with capture_stdout() as result:
      yield
    if not(isinstance(included,list)):
      included = [included]
    if not(isinstance(excluded,list)):
      excluded = [excluded]
    for e in included:
      self.assertTrue(e in result[0],msg=result[0] + "<->" + e)
    for e in excluded:
      self.assertFalse(e in result[0],msg=result[0] + "<->" + e)
      
  @contextmanager
  def assertOutputs(self,included_stdout,included_stderr):
    with capture_stdout() as result:
      yield
    if not(isinstance(included_stdout,list)):
      included_stdout = [included_stdout]
    if not(isinstance(included_stderr,list)):
      included_stderr = [included_stderr]
    for e in included_stdout:
      self.assertTrue(e in result[0],msg=result[0] + "<->" + e)
    for e in included_stderr:
      self.assertTrue(e in result[1],msg=result[1] + "<->" + e)
      
  def tearDown(self):
    t = time.time() - self.startTime
    print("deltaT %s: %.3f" % ( self.id(), t))

  def __init__(self,*margs,**kwargs):
    self.startTime = time.time()
    if sys.version_info >= (3, 0):
      fun = getattr(self,margs[0])
      if not hasattr(fun,'tag_memory_heavy'):
        fun.__dict__["tag_memory_heavy"] = False
      if not hasattr(fun,'tag_slow'):
        fun.__dict__["tag_slow"] = False

      if args.ignore_memory_heavy and fun.tag_memory_heavy:
        fun.__dict__["__unittest_skip__"] = True
        fun.__dict__["__unittest_skip_why__"] = "Ignoring memory_heavy tests (--ignore_memory_heavy)"
      if args.ignore_memory_light and not(fun.tag_memory_heavy):
        fun.__dict__["__unittest_skip__"] = True
        fun.__dict__["__unittest_skip_why__"] = "Ignoring memory_light tests (--ignore_memory_light)"

      if not(args.run_slow) and fun.tag_slow:
        fun.__dict__["__unittest_skip__"] = True
        fun.__dict__["__unittest_skip_why__"] = "Ignoring slow tests (--run_slow)"

    else:
      fun = getattr(getattr(self,margs[0]),'im_func')
      if not hasattr(fun,'tag_memory_heavy'):
        fun.tag_memory_heavy = False
      if not hasattr(fun,'tag_slow'):
        fun.tag_slow = False

      if args.ignore_memory_heavy and fun.tag_memory_heavy:
        fun.__unittest_skip__ = True
        fun.__unittest_skip_why__ = "Ignoring memory_heavy tests (--ignore_memory_heavy)"
      if args.ignore_memory_light and not(fun.tag_memory_heavy):
        fun.__unittest_skip__ = True
        fun.__unittest_skip_why__ = "Ignoring memory_light tests (--ignore_memory_light)"

      if not(args.run_slow) and fun.tag_slow:
        fun.__unittest_skip__ = True
        fun.__unittest_skip_why__ = "Ignoring slow tests (--run_slow)"

    unittest.TestCase.__init__(self,*margs,**kwargs)

  def randDM(self,n,m=1,sparsity=1,valuegenerator=lambda : random.normal(0,1),symm=False ):
    if sparsity < 1:
      spp = self.randDM(n,m,sparsity=1,valuegenerator=lambda : random.uniform(0,1) ,symm=symm)
      spm = (spp < sparsity)
      spm = sparsify(spm)
      ret = DM(spm.sparsity(),array([valuegenerator() for i in range(spm.nnz())]))
      if symm:
        return (ret + ret.T)/2
      else:
        return ret
    else:
      ret = DM([valuegenerator() for i in range(n*m)]).reshape((n,m))
      if symm:
        return (ret + ret.T)/2
      else:
        return ret

  def message(self,s):
      print(s)
      sys.stdout.flush()

  def assertAlmostEqual(self,first, second, places=7, msg=""):
      if isnan(first ) and isnan(second): return
      msg+= " %.16e <-> %.16e"  % (first, second)
      n =  max(abs(first),abs(second))
      if n>1e3:
        n = 10**floor(log10(n))
      else:
        n = 1.0
      unittest.TestCase.assertAlmostEqual(self,first/n,second/n,places=places,msg=msg + "  scaled by %d" %n)

  def checkarray(self,zr,zt,name="",failmessage="",digits=10):
      """
      Checks for equality of two numpy matrices.
      The check uses dense form.

      zr - reference
      zt - to test

      name - a descriptor that will be included in error messages

      """
      if isinstance(zr,tuple):
        zr=array([list(zr)])
      if isinstance(zt,tuple):
        zt=array([list(zt)])
      if isinstance(zr,list):
        zr=array([zr])
      if isinstance(zt,list):
        zt=array([zt])
      if not(hasattr(zt,'shape')) or len(zt.shape)==0:
        zt=array([[zt]])
      if not(hasattr(zr,'shape')) or len(zr.shape)==0:
        zr=array([[zr]])
      if len(zr.shape)==1:
        zr=array(zr,ndmin=2).T
      if len(zt.shape)==1:
        zt=array(zt,ndmin=2).T


      self.assertEqual(zt.shape[0],zr.shape[0],"In %s: %s dimension error. Got %s, expected %s. %s <-> %s" % (name,failmessage,str(zt.shape),str(zr.shape),str(zt),str(zr)))
      self.assertEqual(len(zt.shape),len(zr.shape),"In %s: %s dimension error. Got %s, expected %s. %s <-> %s" % (name,failmessage,str(zt.shape),str(zr.shape),str(zt),str(zr)))
      self.assertEqual(zt.shape[1],zr.shape[1],"In %s: %s dimension error. Got %s, expected %s. %s <-> %s" % (name,failmessage,str(zt.shape),str(zr.shape),str(zt),str(zr)))
      for i in range(zr.shape[0]):
        for j in range(zr.shape[1]):
          try:
            float(zt[i,j])
            float(zr[i,j])
          except:
            self.assertTrue(is_equal(zt[i,j],zr[i,j]),LazyString('"Expressions (%s,%s) are not equal \n %s <-> \n %s at elem(%d,%d): %s <-> %s" % (type(zt),type(zr),str(zt),str(zr),i,j,str(zt[i,j]),str(zr[i,j]))'))
            #self.assertTrue(is_equal(zt[i,j],zr[i,j]),"Expressions (%s,%s) are not equal \n %s <-> \n %s at elem(%d,%d): %s <-> %s" % (type(zt),type(zr),str(zt),str(zr),i,j,str(zt[i,j]),str(zr[i,j])))
            continue
          if zt[i,j]==zr[i,j]:
            continue
          #if (isnan(zt[i,j]) or isinf(zt[i,j])) and  (isinf(zt[i,j]) or isnan(zt[i,j])):
          #  continue
          self.assertAlmostEqual(zt[i,j],zr[i,j],digits,LazyString('"In %s: %s evaluation error.\n %s <->\n %s\n [digits=%d] at elem(%d,%d): " % (name,failmessage,str(zt),str(zr), digits, i,j, )'))
          #self.assertAlmostEqual(zt[i,j],zr[i,j],digits,"In %s: %s evaluation error.\n %s <->\n %s\n [digits=%d] at elem(%d,%d): " % (name,failmessage,str(zt),str(zr), digits, i,j, ))

  def evaluationCheck(self,yt,yr,x,x0,name="",failmessage="",fmod=None,setx0=None):
    """ General unit test for checking casadi evaluation against a reference solution.

        Checks if yr is the same as Function(name,x,yt) , evaluated for x0

        x: the symbolic seed for the test. It should be of a form accepted as first argument of Function
        yt: the test expression.
        yr: the reference solution: a numpy matrix.

        name - a descriptor that will be included in error messages
    """
    self.message(":"+ name)
    if (type(x)==list):
      sample = x[0]
    else :
      sample = x


    if isinstance(sample,SX):
      f = Function("f", x, yt)
    else:
      f = Function("f", x, yt)

    if (not(fmod is None)):
      f=fmod(f,x)
    if not(type(x0)==list):
      x0=[x0]

    if setx0 is None:
      setx0=x0

    if not(type(setx0)==list):
      setx0=[setx0]

    f_in = [0]*f.n_in()
    for i in range(len(x0)):
      try:
        f_in[i]=setx0[i]
      except Exception as e:
         print(f.size_in(i))
         raise e
         raise Exception("ERROR! Tried to set input with %s which is of type  %s \n%s" %(str(x0[i]), str(type(x0[i])),name))
    f_out = f.call(f_in)
    zt = f_out[0].full()
    self.checkarray(yr,zt,name,failmessage)

  def numpyEvaluationCheck(self,ft,fr,x,x0,name="",failmessage="",fmod=None,setx0=None):
    """ General unit test for checking casadi versus numpy evaluation.

        Checks if 'fr(x0)' yields the same as Function(name,x,[ft(x)]) , evaluated for x0

        x: the symbolic seed for the test. It should be of a form accepted as first argument of Function
        ft: the test function. This function should operate on the casadi matrix x and return MX or SX.
        fr: the reference function. This function works on the array x0.

        name - a descriptor that will be included in error messages
    """
    self.message(":"+ name)
    function=None
    frx=None
    try:
      function=ft(x)
      frx=fr(x0)
    except Exception as e:
      print("Error calling functions in %s" % name)
      raise e
    self.evaluationCheck([function],frx,x,x0,name,failmessage,fmod=fmod,setx0=setx0)


  def numpyEvaluationCheckPool(self,pool,x,x0,name="",fmod=None,setx0=None,excludeflags=set()):
    """ Performs a numpyEvaluationCheck for all members of a function pool"""
    for i in range(len(pool.numpyoperators)):
      if len(excludeflags.intersection(pool.flags[i]))>0:
        continue
      self.numpyEvaluationCheck(pool.casadioperators[i],pool.numpyoperators[i],x,x0,"%s:%s" % (name,pool.names[i]),"\n I tried to apply %s (%s) from test case '%s' to numerical value %s. But the result returned: " % (str(pool.casadioperators[i]),pool.names[i],name, str(x0)),fmod=fmod,setx0=setx0)
      
  def check_eval_mx(self, mx):
    if isinstance(mx,list):
        return self.check_eval_mx(vvcat(mx))
    assert isinstance(mx,MX)
    args = symvar(mx)
    f = Function("f",args,[mx])
    d = MX.sym("d") # dummy
    #mx_eval = f.call(args,True,False)[0]
    new_args = [MX.sym(e.name(),e.sparsity()) for e in args]
    mx_eval = f.call(new_args,True,False)[0]
    if isinstance(mx_eval,DM):
        mx = evalf(mx)
        self.checkarray(mx,mx_eval)
        return
    mx_eval = substitute([mx_eval],new_args,args)[0]
    if isinstance(mx_eval,MX): # In all sparse case, might falsely look like SX
        self.check_identical_mx(mx_eval,mx)

  def check_identical_fun(self, a, b):
    s = a.serialize()
    s2 = b.serialize()
    a.save('a.casadi',{"debug":True})
    b.save('b.casadi',{"debug":True})
    self.assertTrue(s==s2)
        
  def check_identical_mx(self, a, b):
    assert isinstance(a,MX)
    assert isinstance(b,MX)
    arg_a = symvar(a)
    arg_b = symvar(b)
    fa = Function('f',arg_a,[a])
    fb = Function('f',arg_b,[b])
    self.check_identical_fun(fa,fb)    

  def checkfunction_light(self,trial,solution,inputs=None,**kwargs):
    self.checkfunction(trial,solution,inputs,fwd=False,adj=False,jacobian=False,gradient=False,hessian=False,sens_der=False,evals=False,**kwargs)
  def checkfunction(self,trial,solution,inputs=None,fwd=True,adj=True,jacobian=True,gradient=True,hessian=True,sens_der=True,evals=True,digits=9,digits_sens=None,failmessage="",allow_empty=True,verbose=True,indirect=False,sparsity_mod=True,allow_nondiff=False):

    if isinstance(inputs,dict):
      d = inputs
      inputs = [0]*trial.n_in()
      ns = trial.name_in()
      for k,v in list(d.items()):
        inputs[ns.index(k)] = v

    if indirect:
      ins = trial.mx_in()
      extra_trial = Function("extra_trial", ins,trial(ins))
      self.checkfunction(extra_trial,solution,fwd,adj,jacobian,gradient,hessian,sens_der,evals,inputs=inputs,digits=digits,digits_sens=digits_sens,failmessage=failmessage,allow_empty=allow_empty,verbose=verbose,indirect=False)

    if digits_sens is None:
      digits_sens = digits

    for i in range(trial.n_in()):
      if (allow_empty and (trial.sparsity_in(i).is_empty() or solution.sparsity_in(i).is_empty() )): continue
      message = "input(%d: '%s')" % (i, trial.name_in(i))

    for i in range(2): # repeated evaluation
      try:
        trial_outputs = trial.call(inputs)
        solution_outputs = solution.call(inputs)
      except Exception as e:
        raise Exception(str(e) + "\nThis occured for simple evaluate(%d,%d) for: %s" % (0,0,failmessage) )

      self.assertEqual(trial.n_out(),solution.n_out(),failmessage+": trial has %d outputs while solution has %d." % (trial.n_out(),solution.n_out()) )
      self.assertEqual(trial.n_in(),solution.n_in(),failmessage+": trial has %d inputs while solution has %d." % (trial.n_in(),solution.n_in()) )

      for i in range(trial.n_out()):
        message = "output(%d: '%s')" % (i, trial.name_out(i))
        if (allow_empty and (trial.sparsity_out(i).is_empty() or solution.sparsity_out(i).is_empty() )): continue
        if (allow_nondiff and (trial.sparsity_out(i).nnz()==0 or solution.sparsity_out(i).nnz()==0 )): continue
        self.checkarray(trial_outputs[i],solution_outputs[i],"",digits=digits,failmessage=failmessage+": "+message)

    if jacobian:
      for i in range(trial.n_in()):
        if (allow_empty and (trial.sparsity_in(i).is_empty() or solution.sparsity_in(i).is_empty() )): continue
        for j in range(trial.n_out()):
          trialjac = jacobian_old(trial, i, j)
          self.assertEqual(trialjac.n_in(),trial.n_in())
          self.assertEqual(trialjac.n_out(),trial.n_out()+1)
          solutionjac = jacobian_old(solution, i, j)
          self.assertEqual(solutionjac.n_in(),solution.n_in())
          self.assertEqual(solutionjac.n_out(),solution.n_out()+1)

          self.checkfunction(trialjac,solutionjac,inputs=inputs,fwd=fwd if sens_der else False,adj=adj if sens_der else False,jacobian=False,gradient=False,hessian=False,evals=False,digits=digits_sens,failmessage="(%s).jacobian_old(%d,%d)" % (failmessage,i,j),allow_empty=allow_empty,verbose=verbose,allow_nondiff=allow_nondiff)

    if hessian:
      for i in range(trial.n_in()):
        if (allow_empty and (trial.sparsity_in(i).is_empty() or solution.sparsity_in(i).is_empty() )): continue
        for j in range(trial.n_out()):
          if trial.sparsity_out(j).is_scalar() and solution.sparsity_out(j).is_scalar():
            trialhess = hessian_old(trial, i, j)
            self.assertEqual(trialhess.n_in(),trial.n_in())
            self.assertEqual(trialhess.n_out(),trial.n_out()+2)
            solutionhess = hessian_old(solution, i, j)
            self.assertEqual(solutionhess.n_in(),solution.n_in())
            self.assertEqual(solutionhess.n_out(),solution.n_out()+2)
            self.checkfunction(trialhess,solutionhess,inputs=inputs,fwd=fwd  if sens_der else False,adj=adj  if sens_der else False,jacobian=False,gradient=False,hessian=False,evals=False,digits=digits_sens,failmessage="(%s).hessian_old(%d,%d)" % (failmessage,i,j),allow_empty=allow_empty,verbose=verbose,allow_nondiff=allow_nondiff)

    if evals is True:
      evals = 2

    if evals:

      def remove_first(x):
        ret = DM(x)
        if ret.numel()>0:
          ret[0,0] = DM(1,1)
          return ret.sparsity()
        else:
          return ret.sparsity()

      def remove_last(x):
        ret = DM(x)
        if ret.nnz()>0:
          ret[ret.sparsity().row()[-1],ret.sparsity().get_col()[-1]] = DM(1,1)
          return ret.sparsity()
        else:
          return ret.sparsity()

      spmods = [lambda x: x , remove_first, remove_last]
      #spmods = [lambda x: x]
      #spmods = [lambda x: x , remove_first]

      sym = MX.sym

      storage2 = {}
      storage = {}

      def vec(l):
        ret = []
        for i in l:
          ret.extend(i)
        return ret

      vf_reference = None

      for f in [solution,trial]:

        # dense
        for spmod,spmod2 in itertools.product(spmods,repeat=2):
          fseeds = [sym("f",spmod(f.sparsity_in(i))) for i in range(f.n_in())]
          aseeds = [sym("a",spmod2(f.sparsity_out(i)))  for i in range(f.n_out())]
          inputss = [sym("i",f.sparsity_in(i)) for i in range(f.n_in())]
          res = f.call(inputss,not f.is_a("SXFunction"))
          #print res, "sp", [i.sparsity().dim(True) for i in fseeds]
          opts = {"helper_options": {"is_diff_in": f.is_diff_in(), "is_diff_out": f.is_diff_out()}}
          [fwdsens] = forward(res, inputss, [fseeds],opts)
          [adjsens] = reverse(res, inputss, [aseeds],opts)

          vf = Function("vf", inputss+vec([fseeds+aseeds]),list(res) + vec([list(fwdsens)+list(adjsens)]),{"is_diff_in": f.is_diff_in()+f.is_diff_in()+f.is_diff_out(), "is_diff_out": f.is_diff_out()+f.is_diff_out()+f.is_diff_in()})

          vf_in = list(inputs)
          # Complete random seeding
          for i in range(f.n_in(),vf.n_in()):
            random.seed(i)
            vf_in.append(DM(vf.sparsity_in(i),random.random(vf.nnz_in(i))))

          vf_out = vf.call(vf_in)
          storagekey = (spmod,spmod2)
          if not(storagekey in storage):
            storage[storagekey] = []
          storage[storagekey].append(vf_out)

          if vf_reference is None:
            vf_reference = vf

          if evals>1:

            # Second order sensitivities
            for spmod_2,spmod2_2 in itertools.product(spmods,repeat=2):
              fseeds2 = [sym("f",vf_reference.sparsity_in(i)) for i in range(vf.n_in())]
              aseeds2 = [sym("a",vf_reference.sparsity_out(i))  for i in range(vf.n_out()) ]
              inputss2 = [sym("i",vf_reference.sparsity_in(i)) for i in range(vf.n_in())]

              res2 = vf.call(inputss2)
              opts = {"helper_options": {"is_diff_in": vf.is_diff_in(), "is_diff_out": vf.is_diff_out()}}
              [fwdsens2] = forward(res2, inputss2, [fseeds2],opts)
              [adjsens2] = reverse(res2, inputss2, [aseeds2],opts)

              vf2 = Function("vf2", inputss2+vec([fseeds2+aseeds2]),list(res2) + vec([list(fwdsens2)+list(adjsens2)]),{"is_diff_in": vf.is_diff_in()+vf.is_diff_in()+vf.is_diff_out(), "is_diff_out": vf.is_diff_out()+vf.is_diff_out()+vf.is_diff_in()})

              vf2_in = list(inputs)

              for i in range(f.n_in(),vf2.n_in()):
                random.seed(i)
                vf2_in.append(DM(vf2.sparsity_in(i),random.random(vf2.nnz_in(i))))

              vf2_out = vf2.call(vf2_in)
              storagekey = (spmod,spmod2)
              if not(storagekey in storage2):
                storage2[storagekey] = []
              storage2[storagekey].append(vf2_out)

      # Remainder of eval testing
      for store,order in [(storage,"first-order"),(storage2,"second-order")][:evals]:
        for stk,st in list(store.items()):
          for i in range(len(st)-1):
            for k,(a,b) in enumerate(zip(st[0],st[i+1])):
              if b.numel()==0 and sparsify(a).nnz()==0: continue
              if a.numel()==0 and sparsify(b).nnz()==0: continue

              if (allow_nondiff and (a.nnz()==0 or b.nnz()==0 )): continue
              #self.checkarray(IM(a.sparsity(),1),IM(b.sparsity(),1),("%s, output(%d)" % (order,k))+str(vf.getInput(0))+failmessage,digits=digits_sens)
              self.checkarray(a,b,("%s, output(%d)" % (order,k))+failmessage,digits=digits_sens)

  def check_sparsity(self, a,b):
    self.assertTrue(a==b, msg=str(a) + " <-> " + str(b))

  def compile_external(self,name,source,opts=None,std="c89",extralibs="",check_serialize=False,extra_options=None,main=False,definitions=None,extra_include=[],debug_mode=False):
    import subprocess
    if args.run_slow:
      libdir = GlobalOptions.getCasadiPath()
      includedir = GlobalOptions.getCasadiIncludePath()
      includedirs = [includedir,os.path.join(includedir,"highs")]
      for e in extra_include:
        includedirs.append(os.path.join(includedir,e))

      if isinstance(extralibs,list):
        extralibs_clean = []
        for lib in extralibs:
            if os.name=='nt':
                if "." in lib:
                    if lib.endswith(".dll"):
                        extralibs_clean.append(lib[:-4]+".lib")
                    else:
                        extralibs_clean.append(lib)
                else:
                    extralibs_clean.append(lib+".lib")
            else:
                if "." in lib:
                    extralibs_clean.append(lib)
                else:
                    extralibs_clean.append("-l"+lib)
        extralibs = " " + " ".join(extralibs_clean)

      if isinstance(extra_options,bool) or extra_options is None:
        extra_options = ""
      if isinstance(extra_options,list):
        extra_options = " " + " ".join(extra_options)
      if definitions is None:
        definitions = []

      def get_commands(shared=True):
        if os.name=='nt':
          defs = " ".join(["/D"+d for d in definitions])
          commands = "cl.exe {shared} {definitions} /D_UCRT_NOISY_NAN {includedir} {source} {extra} /link  /libpath:{libdir} /out:{name}.dll /implib:{name}.lib".format(shared="/LD" if shared else "",std=std,source=source,libdir=libdir,includedir=" ".join(["/I" + e for e in includedirs]),extra=extralibs + extra_options + extralibs + extra_options,definitions=defs,name=name)
          if shared:
            output = "./" + name + ".dll"
          else:
            output = name + ".exe"
          return [commands, output]
        else:
          defs = " ".join(["-D"+d for d in definitions])
          output = "./" + name + (".so" if shared else "")
          flags = "-O3"
          if debug_mode:
            flags = "-O0 -g"
          commands = "gcc -pedantic -std={std} -fPIC {shared} -Wall -Werror -Wextra {includedir} -Wno-dangling-pointer -Wno-unknown-warning-option -Wno-unknown-pragmas -Wno-long-long -Wno-unused-parameter {flags} {definitions} {source} -o {name_out} -L{libdir} -Wl,-rpath,{libdir} -Wl,-rpath,.".format(shared="-shared" if shared else "",std=std,source=source,name_out=name+(".so" if shared else ""),flags=flags,libdir=libdir,includedir=" ".join(["-I" + e for e in includedirs]),definitions=defs) + (" -lm" if not shared else "") + extralibs + extra_options
          if sys.platform=="darwin":
            commands+= " -Xlinker -rpath -Xlinker {libdir}".format(libdir=libdir)
            commands+= " -Xlinker -rpath -Xlinker .".format(libdir=libdir)
          return [commands, output]

      [commands, libname] = get_commands(shared=True)

      print("compile library",commands)
      p = subprocess.Popen(commands,shell=True).wait()
      if sys.platform=="darwin":
        subprocess.run(["otool","-l",libname])
      if opts is None: opts = {}
      return (external(name, libname,opts),libname)

  def check_codegen(self,F,inputs=None, opts=None,std="c89",extralibs="",check_serialize=False,extra_options=None,main=False,main_return_code=0,definitions=None,with_jac_sparsity=False,external_opts=None,with_reverse=False,with_forward=False,extra_include=[],digits=15,debug_mode=False):
    if not isinstance(main_return_code,list):
        main_return_code = [main_return_code]
    if args.run_slow:
      import hashlib
      name = "codegen_%s" % (hashlib.md5(("%f" % np.random.random()+str(F)+str(time.time())).encode()).hexdigest())
      if opts is None: opts = {}
      if main: opts["main"] = True
      cg = CodeGenerator(name,opts)
      cg.add(F,with_jac_sparsity)
      if with_reverse:
        cg.add(F.reverse(1), with_jac_sparsity)
      if with_forward:
        cg.add(F.forward(1), with_jac_sparsity)
      cg.generate()
      import subprocess

      libdir = GlobalOptions.getCasadiPath()
      includedir = GlobalOptions.getCasadiIncludePath()
      includedirs = [includedir,os.path.join(includedir,"highs")]
      for e in extra_include:
        includedirs.append(os.path.join(includedir,e))

      if isinstance(extralibs,list):
        extralibs_clean = []
        for lib in extralibs:
            if os.name=='nt':
                if "." in lib:
                    if lib.endswith(".dll"):
                        extralibs_clean.append(lib[:-4]+".lib")
                    else:
                        extralibs_clean.append(lib)
                else:
                    extralibs_clean.append(lib+".lib")
            else:
                if "." in lib:
                    extralibs_clean.append(lib)
                else:
                    extralibs_clean.append("-l"+lib)
        extralibs = " " + " ".join(extralibs_clean)

      if isinstance(extra_options,bool) or extra_options is None:
        extra_options = ""
      if isinstance(extra_options,list):
        extra_options = " " + " ".join(extra_options)
      if definitions is None:
        definitions = []

      def get_commands(shared=True):
        if os.name=='nt':
          defs = " ".join(["/D"+d for d in definitions])
          commands = "cl.exe {shared} {definitions} /D_UCRT_NOISY_NAN {includedir} {name}.c {extra} /link  /libpath:{libdir}".format(shared="/LD" if shared else "",std=std,name=name,libdir=libdir,includedir=" ".join(["/I" + e for e in includedirs]),extra=extralibs + extra_options + extralibs + extra_options,definitions=defs)
          if shared:
            output = "./" + name + ".dll"
          else:
            output = name + ".exe"
          return [commands, output]
        else:
          defs = " ".join(["-D"+d for d in definitions])
          output = "./" + name + (".so" if shared else "")
          flags = "-O3"
          if debug_mode:
            flags = "-O0 -g"
          commands = "gcc -pedantic -std={std} -fPIC {shared} -Wall -Werror -Wextra {includedir} -Wno-dangling-pointer -Wno-unknown-warning-option -Wno-unknown-pragmas -Wno-long-long -Wno-unused-parameter {flags} {definitions} {name}.c -o {name_out} -L{libdir} -Wl,-rpath,{libdir} -Wl,-rpath,.".format(shared="-shared" if shared else "",std=std,name=name,name_out=name+(".so" if shared else ""),libdir=libdir,includedir=" ".join(["-I" + e for e in includedirs]),definitions=defs,flags=flags) + (" -lm" if not shared else "") + extralibs + extra_options
          if sys.platform=="darwin":
            commands+= " -Xlinker -rpath -Xlinker {libdir}".format(libdir=libdir)
            commands+= " -Xlinker -rpath -Xlinker .".format(libdir=libdir)
          return [commands, output]

      [commands, libname] = get_commands(shared=True)

      print("compile library",commands)
      p = subprocess.Popen(commands,shell=True).wait()
      #if sys.platform=="darwin":
      #  subprocess.run(["otool","-l",libname])
      if external_opts is None: external_opts = {}
      F2 = external(F.name(), libname,external_opts)

      if main:
        [commands, exename] = get_commands(shared=False)
        print("here",commands)
        env = os.environ
        if os.name=='nt':
            env["PATH"] = env["PATH"]+";"+libdir
        else:
            env["LD_LIBRARY_PATH"] = libdir
        p = subprocess.Popen(commands,shell=True,env=env).wait()
        inputs_main = inputs
        if isinstance(inputs,dict):
          inputs_main = F.convert_in(inputs)
        F.generate_in(F.name()+"_in.txt", inputs_main)

      if main:
        with open(F.name()+"_out.txt","w") as stdout:
          with open(F.name()+"_in.txt","r") as stdin:
            commands = exename+" "+F.name()
            print(commands+" < " + F.name()+"_in.txt")
            p = subprocess.Popen(commands,shell=True,stdin=stdin,stdout=stdout)
            out = p.communicate()
        print("Return code",p.returncode)
        assert p.returncode in main_return_code
        if main_return_code:
            return
        
      Fout = F.call(inputs)
      Fout2 = F2.call(inputs)

      if main:
        outputs = F.generate_out(F.name()+"_out.txt")
        print(outputs)
        if isinstance(inputs,dict):
          outputs = F.convert_out(outputs)
          for k in F.name_out():
            self.checkarray(Fout[k],outputs[k],digits=digits)
        else:
          for i in range(F.n_out()):
            self.checkarray(Fout[i],Fout2[i],digits=digits)

      if isinstance(inputs, dict):
        self.assertEqual(F.name_out(), F2.name_out())
        for k in F.name_out():
          self.checkarray(Fout[k],Fout2[k],digits=digits,failmessage=k)
      else:
        for i in range(F.n_out()):
          self.checkarray(Fout[i],Fout2[i],digits=digits)

      if self.check_serialize:
        self.check_serialize(F2,inputs=inputs)
        
      return F2, libname

  def check_thread_safety(self,F,inputs=None,N=20):

    FP = F.map(N, 'thread',2)
    FS = F.map(N, 'thread')
    self.checkfunction_light(FP, FS, inputs)


  def check_serialize(self,F,inputs=None):
      F2 = Function.deserialize(F.serialize({"debug":True}))

      Fout = F.call(inputs)
      Fout2 = F2.call(inputs)

      if isinstance(inputs, dict):
        self.assertEqual(F.name_out(), F2.name_out())
        for k in F.name_out():
          self.checkarray(Fout[k],Fout2[k],digits=16)
      else:
        for i in range(F.n_out()):
          self.checkarray(Fout[i],Fout2[i],digits=16)


      if False:
          import hashlib
          h = hashlib.md5(F.serialize({"debug":True}).encode("ascii")).hexdigest()
          path = "serialize_"+casadi.__version__
          os.makedirs(path, exist_ok=True)
          F.save(os.path.join(path,h+".casadi"),{"debug":True})
          DM.set_precision(16)
          h_in = hashlib.md5(str(inputs).encode("ascii")).hexdigest()
          DM.set_precision(7)
          if isinstance(inputs,dict):
            inputs = F.convert_in(inputs)
            Fout = F.convert_out(Fout)
          F.generate_in(os.path.join(path,h+"_"+h_in+"_in.txt"),inputs)
          F.generate_out(os.path.join(path,h+"_"+ h_in +"_out.txt"),Fout)


  def check_pure(self,F,inputs=None):
      Fout = F.call(inputs)
      Fout2 = F.call(inputs)

      if isinstance(inputs, dict):
        for k in F.name_out():
          self.checkarray(Fout[k],Fout2[k],digits=16)
      else:
        for i in range(F.n_out()):
          self.checkarray(Fout[i],Fout2[i],digits=16)

class run_only(object):
  def __init__(self, args):
    self.args = []
    for a in args:
      self.args.append(a)

  def __call__(self, c):
    print("run_only:")
    for i in dir(c):
      if i.startswith('test_'):
        n = i[5:]
        if not n in self.args:
          delattr(c,i)
        else:
          print(i)
    return c

class requires(object):
  def __init__(self,att):
    self.att = att

  def __call__(self,c):
    if hasattr(casadi,self.att):
      return c
    else:
      print("Not available %s, skipping unittests" % self.att)
      return None

class requires_conic(object):
  def __init__(self,n):
    self.n = n

  def __call__(self,c):
    try:
      load_conic(self.n)
      return c
    except:
      print("Not available QP plugin %s, skipping unittests" % self.n)
      return None

class requires_linsol(object):
  def __init__(self,n):
    self.n = n

  def __call__(self,c):
    try:
      load_linsol(self.n)
      return c
    except:
      print("Not available linsol plugin %s, skipping unittests" % self.n)
      return None

class requires_nlpsol(object):
  def __init__(self,n):
    self.n = n

  def __call__(self,c):
    import os
    if "SKIP_" + self.n.upper() + "_TESTS" in os.environ:
        return None
    try:
      load_nlpsol(self.n)
      return c
    except:
      print("Not available NLP plugin %s, skipping unittests" % self.n)
      return None

class requires_expm(object):
  def __init__(self,n):
    self.n = n

  def __call__(self,c):
    try:
      load_expm(self.n)
      return c
    except:
      print("Not available Expm plugin %s, skipping unittests" % self.n)
      return None

class requires_integrator(object):
  def __init__(self,n):
    self.n = n

  def __call__(self,c):
    try:
      load_integrator(self.n)
      return c
    except:
      print("Not available integrator plugin %s, skipping unittests" % self.n)
      return None

class requires_rootfinder(object):
  def __init__(self,n):
    self.n = n

  def __call__(self,c):
    try:
      load_rootfinder(self.n)
      return c
    except:
      print("Not available RFP plugin %s, skipping unittests" % self.n)
      return None

class requiresPlugin(object):
  def __init__(self,att,n):
    self.att = att
    self.n = n

  def __call__(self,c):
    try:
      self.att.load_plugin(self.n)
      return c
    except:
      print("Not available %s plugin %s, skipping unittests" % (str(self.att),self.n))
      return None

class skip(object):
  def __init__(self, skip=True):
    self.skip = skip

  def __call__(self, c):
    if not self.skip: return c
    if isinstance(c,unittest.TestCase):
      if i.startswith('test_'):
        delattr(c,i)
      return c
    else:
      print(self.skiptext(c.__name__))
      return None

  def skiptext(self,name):
    return "Skipping test '%s'" % name

def known_bug():
  return unittest.skipIf(not(args.known_bugs),"known bug (run with --known_bugs to include it)")

class memory_heavy(object):
  def __init__(self):
    pass

  def __call__(self, c):
    print(c)
    if c is not None:
        c.__dict__["tag_memory_heavy"] = True
    return c

class slow(object):
  def __init__(self):
    pass

  def __call__(self, c):
    print("slow", c)
    c.__dict__["tag_slow"] = True
    return c



# This is an automatically generated file converting from the apm format

import casadi as cs

def hs005():
    # The optimal objective is (if given in):
    x_opt = cs.DM([-cs.pi/3 + 0.5, -cs.pi/3 - 0.5])
    f_opt = -(cs.sqrt(3)/2 + cs.pi/3)
    x = cs.MX.sym('x', 2)
    obj = cs.MX.zeros(1, 1)
    x0 = cs.DM.zeros(2, 1)
    lbx = -cs.inf*cs.DM.ones(2, 1)
    ubx = cs.inf*cs.DM.ones(2, 1)
    g = []
    lbg = []
    ubg = []
    
    x0[0]= 0
    lbx[0]= -1.5
    ubx[0]= 4
    x0[1]= 0
    lbx[1]= -3
    ubx[1]=3
    
    obj = cs.sin(x[0]+x[1]) + (x[0]-x[1])**2 - 1.5*x[0] + 2.5*x[1] + 1
    
    
    f = obj
    
    return (x_opt, f_opt, x, f, g, lbg, ubg, lbx, ubx, x0)



#
#     MIT No Attribution
#
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl, KU Leuven.
#
#     Permission is hereby granted, free of charge, to any person obtaining a copy of this
#     software and associated documentation files (the "Software"), to deal in the Software
#     without restriction, including without limitation the rights to use, copy, modify,
#     merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
#     permit persons to whom the Software is furnished to do so.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
#     INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
#     PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#     OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# -*- coding: utf-8 -*-
"""
We want to model a chain attached to two supports and hanging in between. Let us discretise
it with N mass points connected by N-1 springs. Each mass i has position (yi,zi), i=1,...,N.
The equilibrium point of the system minimises the potential energy.

The potential energy of each spring is
Vi=D_i/2 * ((y_i-y_{i+1})^2 + (z_i-z_{i+1})^2)

The gravitational potential energy of each mass is
Vg_i = m_i*g0*z_i

The total potential energy is thus given by:
 
Vchain(y,z) = 1/2*sum{i=1,...,N-1} D_i ((y_i-y_{i+1})^2+(z_i-z_{i+1})^2) + g0 * sum{i=1,...,N} m_i * z_i

where y=[y_1,...,y_N] and z=[z_1,...,z_N]

We wish to solve
minimize{y,z} Vchain(y, z)

Subject to the piecewise linear ground constraints:
z_i >= zin
z_i - 0.1*y_i >= 0.5
"""

from casadi import *

# Constants
N = 40
m_i = 40.0/N
D_i = 70.0*N
g0 = 9.81
#zmin = -inf # unbounded
zmin = 0.5 # ground

# Objective function
Vchain = 0

# Variables
x = []

# Variable bounds
lbx = []
ubx = []

# Constraints
g = []

# Constraint bounds
lbg = []
ubg = []

# Loop over all chain elements
for i in range(1, N+1):
   # Previous point
   if i>1:
      y_prev = y_i
      z_prev = z_i

   # Create variables for the (y_i, z_i) coordinates
   y_i = SX.sym('y_' + str(i))
   z_i = SX.sym('z_' + str(i))

   # Add to the list of variables
   x += [y_i, z_i]
   if i==1:
    lbx += [-2., 1.]
    ubx += [-2., 1.]
   elif i==N:
    lbx += [ 2., 1.]
    ubx += [ 2., 1.]
   else:
    lbx += [-inf, zmin]
    ubx += [ inf,  inf]

   # Spring potential
   if i>1:
      Vchain += D_i/2*((y_prev-y_i)**2 + (z_prev-z_i)**2)

   # Graviational potential
   Vchain += g0 * m_i * z_i

   # Slanted ground constraints
   g.append(z_i - 0.1*y_i)
   lbg.append( 0.5)
   ubg.append( inf)

# Formulate QP
qp = {'x':vertcat(*x), 'f':Vchain, 'g':vertcat(*g)}

# Solve with IPOPT
solver = qpsol('solver', 'qpoases', qp, {'sparse':True})
#solver = qpsol('solver', 'gurobi', qp)
#solver = nlpsol('solver', 'ipopt', qp)

# Get the optimal solution
sol = solver(lbx=lbx, ubx=ubx, lbg=lbg, ubg=ubg)
x_opt = sol['x']
f_opt = sol['f']
print('f_opt = ', f_opt)

# Retrieve the result
Y0 = x_opt[0::2]
Z0 = x_opt[1::2]

# Plot the result
import matplotlib.pyplot as plt
plt.plot(Y0,Z0,'o-')
ys = linspace(-2.,2.,100)
zs = 0.5 + 0.1*ys
plt.plot(ys,zs,'--')
plt.xlabel('y [m]')
plt.ylabel('z [m]')
plt.title('hanging chain QP')
plt.grid(True)
plt.legend(['Chain','z - 0.1y >= 0.5'],loc=9)
plt.show()



#
#     This file is part of CasADi.
#
#     CasADi -- A symbolic framework for dynamic optimization.
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl,
#                             KU Leuven. All rights reserved.
#     Copyright (C) 2011-2014 Greg Horn
#
#     CasADi is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     CasADi is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#LH9JCUD52HXU
from casadi import *
import casadi as c
import numpy
import unittest
from types import *
from helpers import *
import random
from collections import defaultdict
import sys


class SerializeTests(casadiTestCase):


  def test_compat(self):
  
    dir = "serialize_3.5.5"
    if not os.path.isdir(dir):
        return
    
    errors = {}
    
    inputs = defaultdict(list)
    outputs = defaultdict(list)
    names = os.listdir(dir)
    functions = [n for n in names if n.endswith(".casadi")]
    for n in names:
        data = inputs if "_in" in n else outputs
        data[n.split("_")[0]].append(n)
    for fun in functions:
      try:
          f = Function.load(os.path.join(dir,fun))
      except Exception as e:
          if "DllLibrary::init_handle" in str(e):
            continue
          if "DeSerialization of Integrator failed" in str(e):
            continue
          if "DeSerialization of Rootfinder failed" in str(e):
            continue
          else:
             raise Exception(str(e))

      key = fun.split(".")[0]
      for in_name,out_name in zip(sorted(inputs[key]),sorted(outputs[key])):
          inp = f.generate_in(os.path.join(dir,in_name))
          outp_ref = f.generate_out(os.path.join(dir,out_name))
          outp = f.call(inp)
          self.assertEqual(len(outp),len(outp_ref))
          for o,o_ref in zip(outp,outp_ref):
            digits = 15
            if "SuperscsInterface" in str(f):
                digits = 7
            if sys.platform=="darwin":
                digits = 7 # Bug?
            if sys.platform=="win32":
                digits = 7 # Bug?
            self.checkarray(o,o_ref,digits=digits,failmessage=fun+"/"+str(f))
            
  def test_identity(self):
    obj = ["foo",{"foo":"bar"},[{"a":3},{"b":9}],["a",5],{"foo": ["a",5]},{"foo": [["a",5],["b",2]]},[["a",5],["b",2]],[[["a",5],["b",2]]]]
    
    def check_equal(a,b):
        if isinstance(a,dict):
            assert list(sorted(a.keys()))==list(sorted(b.keys()))
            for k in a.keys():
                check_equal(a[k],b[k])
        elif isinstance(a,list):
            assert(len(a)==len(b))
            for i in range(len(a)):
                check_equal(a[i],b[i])
        else:
            assert a==b
        
    for e in obj:
        ss = StringSerializer()
        ss.pack(e)
        
        ds = StringDeserializer(ss.encode())
        r = ds.unpack()
        
        print(e,r)
        check_equal(e,r)
      
if __name__ == '__main__':
    unittest.main()



from casadi import *

n = 5
colind = [0, 3, 6, 8, 10, 12]
row = [0, 1, 4, 1, 2, 4, 0, 2, 0, 3, 3, 4]
nz = [19, 12, 12, 21, 12, 12, 21, 16, 21, 5, 21, 18]
Asp = Sparsity(n, n, colind, row)
A0 = DM(Asp, nz)
A0 = mtimes(A0.T, A0)
Asp = A0.sparsity()
b0 = DM.ones(n)
x0 = solve(A0, b0)
print(x0)

# Test sparse QR
[D, LT, p] = ldl(A0)
x0 = ldl_solve(b0, D, LT, p)
print(x0)

A = MX.sym('A', Asp)
b = MX.sym('b', n)
x = solve(A, b, 'ldl')

for jit in [False, True]:
    # Create function
    f = Function('f', [A,b], [x], ['A', 'b'], ['x'], dict(jit=jit))

    # Numerical evaluation
    x0 = f(A0, b0)
    print(x0)



# This is an automatically generated file converting from the apm format

import casadi as cs

def hs003():
    # The optimal objective is (if given in):
    f_opt = 0
    x_opt = cs.DM([0,0])
    x = cs.MX.sym('x', 2)
    obj = cs.MX.zeros(1, 1)
    x0 = cs.DM.zeros(2, 1)
    lbx = -cs.inf*cs.DM.ones(2, 1)
    ubx = cs.inf*cs.DM.ones(2, 1)
    g = []
    lbg = []
    ubg = []
    
    x0[0]= 10
    x0[1]= 1
    lbx[1]=0
    
    obj = x[1] + 0.00001*(x[1]-x[0])**2
    
    
    f = obj
    
    return (x_opt, f_opt, x, f, g, lbg, ubg, lbx, ubx, x0)



#
#     MIT No Attribution
#
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl, KU Leuven.
#
#     Permission is hereby granted, free of charge, to any person obtaining a copy of this
#     software and associated documentation files (the "Software"), to deal in the Software
#     without restriction, including without limitation the rights to use, copy, modify,
#     merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
#     permit persons to whom the Software is furnished to do so.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
#     INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
#     PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#     OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# -*- coding: utf-8 -*-
from casadi import *
import pylab as plt

# Pendulum: point mass on massless rod 
# x'' = T*x, y'' = T*y-g, x**2+y**2-L**2
#
#
#     .
#  L /
#   /
#  /
# o (x,y) 
#
x = SX.sym("x") # horizontal position of point
y = SX.sym("y") # vertical position of point

dx = SX.sym("dx")
dy = SX.sym("dy")

u = SX.sym("u") # helper states to bring second order system to first order
v = SX.sym("v")

du = SX.sym("du")
dv = SX.sym("dv")

# Force
T = SX.sym("T")


L = 1
g = 9.81

alg = vertcat(dx-u,dy-v,du-T*x,dv-T*y+g,x**2+y**2-L**2)

# Implicit differential states
x_impl  = vertcat(x,y,u,v)
# Derivatives of implict differential states
dx_impl = vertcat(dx,dy,du,dv)

z = T

dae = {"x_impl": x_impl, "dx_impl": dx_impl, "z": z, "alg": alg}

# Perform index reduction
(dae_reduced,stats) = dae_reduce_index(dae)
print("We had an index 3 system (reduced to index 1 now)")
print(stats)
print("Here's the reduced DAE:")
print(dae_reduced)
print("Notice how the algebraic equation now contains a second derivative of x**2+y**2-L**2")
print("Also note how x**2+y**2-L**2 and its first derivate are absent from the algebraic equations.")
print("Both are kept in 'I' though: the resulting DAE invariants.")

# The DAE is not yet in a form that CasADi integrator can handle (semi-explicit).
# Let's convert it, ad obtain some adaptors
[dae_se, state_to_orig, phi] = dae_map_semi_expl(dae, dae_reduced)

grid = list(np.linspace(0,5,1000))
tf = DM(grid).T
intg = integrator("intg","idas",dae_se,0,grid)

# Before we can integrate, we need a consistant initial guess
# Let's say we start at y=-0.5, dx=-0.1;
# consistency equations should automatically identify everything else

# Encode the desire to only enforce y and dx
init_strength = {}
free  = 0 # default
force = -1
init_strength["x_impl"] = vertcat(free,force,free,free)
init_strength["dx_impl"] = vertcat(force,free,free,free)
init_strength["z"] = vertcat(free)

# Obtain a generating Function for initial guesses
init_gen = dae_init_gen(dae, dae_reduced, "ipopt", init_strength)

init = {}
blank = 0.0 # Will not be enforced but still used as initial guess
# suggest to initialize with the left-hanging solution by having x=-1 as initial guess 
init["x_impl"]  = vertcat(-1.0,-0.5,blank,blank)  
init["dx_impl"] = vertcat(-0.1,blank,blank,blank)
init["z"]  = blank

xz0 = init_gen(**init)
print("We have a consistent initial guess now to feed to the integrator:")
print(xz0)

print("Look, we found that force in the pendulum at t=0 equals:")
print(state_to_orig(xf=xz0['x0'],zf=xz0['z0'])["z"])

print("A consistent initial guess should make the invariants zero (up to solver precision):")
print(phi(x=xz0['x0'],z=xz0['z0']))

# Integrate and get resultant xf,zf on grid
sol = intg(**xz0)

print(sol['xf'].shape)

# Solution projected into original DAE space
sol_orig = state_to_orig(xf=sol["xf"],zf=sol["zf"])

# We can see the large-angle pendulum motion play out well in the u state
plt.plot(tf.T,sol_orig["x_impl"].T)
plt.grid(True)
plt.legend([e.name() for e in vertsplit(x_impl)])
plt.xlabel("Time [s]")
plt.title("Boundary value problem solution trajectory")

# A perfect integrator will perfectly preserve the values of the invariants over time
# Integrator errors make the invariants drift in practice
# This is not a detail; the pendulum length is in fact growing!
error = phi(x=sol["xf"],z=sol["zf"])["I"]

plt.figure()
plt.plot(tf.T,error.T)
plt.grid(True)
plt.xlabel("Time [s]")
plt.title("Evolution trajectory of invariants")

# There are some techniques to avoid the drifting of invariants associated with index reduction
#  - Method of Dummy Derviatives (not implemented)
#    hard implementation details: may need to switch between different choices online
#     -> integration in parts triggered by zero-crossing events
#  - Stop the integration every once in a while and project back (not implemented)
#  - Baumgarte stabilization (implemented): build into the equations a form of continuous feedback that brings back deviations in invariants back into the origin


# Demonstrate Baumgarte stabilization with a pole of -1.
# Drift is absent now.
(dae_reduced,stats) = dae_reduce_index(dae, {"baumgarte_pole": -1})
[dae_se, state_to_orig, phi] = dae_map_semi_expl(dae, dae_reduced)
intg = integrator("intg","idas",dae_se,0,grid)
init_gen = dae_init_gen(dae, dae_reduced, "ipopt", init_strength)
xz0 = init_gen(**init)
sol = intg(**xz0)
error = phi(x=sol["xf"],z=sol["zf"])["I"]
plt.figure()
plt.plot(tf.T,error.T)
plt.grid(True)
plt.xlabel("Time [s]")
plt.title("Boundary value problem solution trajectory with Baumgarte pole=-1")



# Sweep across different integrator precisions and Baumgarte poles.
# The choice of pole is not really straightforward
plt.figure()

poles = [-0.1,-1,-10]
precisions = [0.1,1,10]

abstol_default = 1e-8
reltol_default = 1e-6

plt.subplot(len(precisions),len(poles),1)

for i,pole in enumerate(poles):
  for j,precision in enumerate(precisions):
    (dae_reduced,stats) = dae_reduce_index(dae, {"baumgarte_pole": pole})
    [dae_se, state_to_orig, phi] = dae_map_semi_expl(dae, dae_reduced)
    intg = integrator("intg","idas",dae_se,0,grid,{"abstol":abstol_default*precision,"reltol":reltol_default*precision})
    init_gen = dae_init_gen(dae, dae_reduced, "ipopt", init_strength, {"ipopt.print_level":0,"print_time": False})
    xz0 = init_gen(**init)
    sol = intg(**xz0)
    nfevals = intg.stats()["nfevals"]
    error = phi(x=sol["xf"],z=sol["zf"])["I"]
    plt.subplot(len(precisions),len(poles),j*len(poles)+i+1)
    plt.plot(tf.T,error.T)
    plt.grid(True)
    plt.xlabel("Time [s]")
    plt.title("pole=%0.1f, abstol=%0.0e => nfevals=%d" % (pole,abstol_default*precision,nfevals))

plt.show()



#
#     This file is part of CasADi.
#
#     CasADi -- A symbolic framework for dynamic optimization.
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl,
#                             KU Leuven. All rights reserved.
#     Copyright (C) 2011-2014 Greg Horn
#
#     CasADi is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     CasADi is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#
"""
This script reads XML doxygen output and reads source-files to harvest Options.

This script generates a header file with option information that doxygen can use.
"""


xml = 'XML_internal/'
out = 'extra/'

import re
import shutil
from lxml import etree
import os
import re
import fnmatch

from generate_options_helpers import addExtra

# We will use an XML parser for the main index XML file
xmlData = etree.parse(xml+'index.xml')

classes = xmlData.findall("//compound[@kind='class']")

# The main data structure we will set up
# Keys are the full names e.g. casadi::Sundials::CVodesIntegrator
# The values are dicts with the following structure:
# 
# parents     - the list of classes from which this class inherits
# hierarchy   - the list of ancestors of this class. First element is the parent.
# file        - the absolute path to the hpp source file
# xmlsource   - the relative path (wrt to xml) to the xml file describing the class
# hasInternal - (optional) the internal class of this class (the relative path to the xml file)
# InternalFor - the list of classes for which this class is internal
# options     - a list of dicts denoting the class's options
# monitors    - a list of dicts denoting the class's monitorables
# stats       - a list of dicts denoting the class's stats
# inputscheme  - the input scheme of the function
# outputscheme - the output scheme of the function
metadata=dict()

# Text-parsing to fill in 'parents','file','hasInternal'
for c in classes:
  refid = c.attrib['refid']
  name=c.findtext("name")

  # Parse the XML file that describes the class
  f = etree.parse(xml+refid+'.xml')
  
  meta = dict()
  metadata[name]=meta
  meta['parents']=[]
  meta['xmlsource']=refid
  
  if f.getroot().find("compounddef/briefdescription/para") is not None:
    meta["brief"] = "".join(list(f.getroot().find("compounddef/briefdescription/para").itertext())).strip()
  else:
    meta["brief"] = ""
 
  # find parents
  for e in f.getroot().findall("compounddef/basecompoundref"):
    if e.text.find("std::")==-1: # Exclude standard library from hierarchy
      meta['parents'].append(e.text)

  # find the internal class of this class, if any
  temp = f.find("//memberdef[name='operator->']/type/ref")
  if not(temp is None):
    meta['hasInternal']=temp.attrib["refid"]
  
  # find the file location of this class
  temp = f.find("compounddef/location")
  if not(temp is None):
    meta['file']=temp.attrib["file"]
    

for v in metadata.values(): # Remove templating if not in index.xml
  for i,vv in enumerate(v['parents']):
    if not(vv in metadata):
      v['parents'][i] = vv.split("<")[0]
      
# Get the parents of a class
def parents(name):
  if name not in metadata:
    return []
  return metadata[name]['parents'] + sum(map(parents,metadata[name]['parents']),[])

# Fill in 'hierarchy'
for k in metadata.keys():
  metadata[k]['hierarchy']=parents(k)

# Get name by specifying xml-source
def nameByXMLsource(xmlsource):
  for name,meta in metadata.items():
    if (meta['xmlsource']==xmlsource):
      return name
  return None

# Fill in 'internalFor'
for name,meta in metadata.items():
  if ('hasInternal' in meta):
    name2 = nameByXMLsource(meta['hasInternal'])
    if name2 is None:
      continue
    internalfor = metadata[name2]
    if ('InternalFor' in internalfor):
      internalfor['InternalFor'].append(name)
    else:
      internalfor['InternalFor']=[name]

# http://code.activestate.com/recipes/499305/
def locate(pattern, root=os.curdir):
    '''Locate all files matching supplied filename pattern in and below
    supplied root directory.'''
    for path, dirs, files in os.walk(os.path.abspath(root)):
        for filename in fnmatch.filter(files, pattern):
            yield os.path.join(path, filename)
  
# locate the namespacecasadi.xml
xmlNS = etree.parse(list(locate('namespacecasadi.xml',root=xml))[0])

# construct a table with enum info
enums = {}

for enum in xmlNS.findall("//memberdef[@kind='enum']"):
  name=enum.findtext("name")
  enums[name]=[]
  description = ""
  
  if enum.findtext("briefdescription/para"):
    description+="".join(list(enum.find("briefdescription/para").itertext())).strip()
  if enum.findtext("detaileddescription/para"):
    description+= "\n" + "".join(list(enum.find("detaileddescription/para").itertext())).strip()
  enums[name].append(description)
  
  for enumvalue in enum.findall("enumvalue"):
    name_=enumvalue.findtext("name")
    description = ""
    if enumvalue.findtext("briefdescription/para"):
      description += enumvalue.findtext("briefdescription/para").strip()
    if enumvalue.findtext("detaileddescription/para"):
      description += "\n" + enumvalue.findtext("detaileddescription/para").strip()
    enums[name].append({'name': name_, 'description': description.strip()})
    
from pyparsing import *


multiline_string = OneOrMore(dblQuotedString)

def removequotes(s):
  return s[1:-1]
  
parse_quoted_string = multiline_string.copy().setParseAction(lambda x: "".join(map(removequotes,x)))
parse_quoted_string_keep = multiline_string.copy().setParseAction(lambda x: '"' + ("".join(map(removequotes,x)))+'"')
parse_type = Word( srange("[A-Z]") + "_").setResultsName("type")

comma = Suppress(Literal(","))

def brackets(a):
  return Suppress(Literal("{")) + Optional(a) + Suppress(Literal("}"))

def gbrackets(a):
  return Group(brackets(a))
  
def delimitedListTrailing( expr, delim=","):
    return ( expr + ZeroOrMore( Suppress( delim ) + expr ) + Optional( Suppress(delim) ) )


parse_optiondata = parse_type + comma + parse_quoted_string
parse_optiondata_collection = gbrackets(parse_optiondata)

parse_option = parse_quoted_string + comma + parse_optiondata_collection
parse_option_collection = gbrackets(parse_option)

parse_options = delimitedListTrailing(parse_option_collection, delim=",")
parse_options_collection = gbrackets(parse_options)

parse_options_group = (brackets( Suppress(Literal("&")) + Word(alphanums + ".:-_"))) | parse_options_collection
parse_options_group_collection = Optional(Suppress(Literal("="))) + brackets(delimitedList(parse_options_group)) + Suppress(Literal(";"))
 
# Inspect anything that has FXInternal as Base Class
for name,meta in metadata.items():
  if not('file' in meta):
    meta['options']={}
    meta['stats']={}
    meta['monitors']={}
    meta['optionproviders'] = []
    continue
  source = re.sub(r'(_impl)?\.hpp$',r'.cpp',meta['file'])
  meta['options']={}
  meta['stats']={}
  meta['monitors']={}
  meta['optionproviders'] = []
  try:
    f =open(source,"r")
  except:
    continue
  lines = f.readlines()
  linec = 0
  while linec<len(lines):
    l = lines[linec]
    linec+=1
    if 'Options' in l:
      if re.search('Options .*::options_',l):
        start = linec
        while "};" not in l:
          l = lines[linec]
          linec+=1
        stop = linec
        
        target = lines[start-1]
        if name.split("::")[1] not in target:
            continue
        optionsdict = "".join(lines[start:stop])

        try:
            results = parse_options_group_collection.parseString(optionsdict)
        except Exception as e:
            print("parse_options_group_collection",parse_options_group_collection)
            print("optionsdict",optionsdict)
            raise e

        for optiongroup in results:
          if isinstance(optiongroup,str):
            m = re.search("(.*?)::options_",optiongroup)
            assert m
            op = m.group(1)
            if "casadi" not in op: op = "casadi::" + op
            meta['optionproviders'].append(op)
            continue
          if len(optiongroup[0])==0:
            continue
          for op_name, op_data in optiongroup:   
            d = meta['options'][op_name]={'name': op_name,"type": op_data[0],'used': name,'default':'','description':op_data[1],'inherit': False}
      
    if not(l.find('ops_')==-1):
      m = re.search(r'ops_\["(.*?)"\]\s*=\s*(.*?);',l)
      if m:
        meta['options'][m.group(1)]={'name': m.group(1),'type': m.group(2),'default': '','description': '', 'used': name}
    if not(l.find('stats_')==-1):
      m = re.search(r'stats_\["(.*?)"\]',l)
      if m:
        meta['stats'][m.group(1)]={'name': m.group(1), 'used': name}
    if not(l.find('INPUTSCHEME')==-1):
      m = re.search(r'INPUTSCHEME\((.*?)\)',l)
      if m:
        meta['inputscheme']=m.group(1)
    if not(l.find('OUTPUTSCHEME')==-1):
      m = re.search(r'OUTPUTSCHEME\((.*?)\)',l)
      if m:
        meta['outputscheme']=m.group(1)

  for k in ['options','stats','monitors']:
    if len(meta[k])==0:
      del meta[k]
  f.close()

addExtra(metadata)

def htmlescape(h):
  return h.replace(">","&gt;").replace("<","&lt;")

def newline2br(a):
  return a.replace("\n","<br />")
  
def optionsashtml(option,used=True):
  if used:
    return "<tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>" %(option['name'],option['type'],newline2br(htmlescape(option['description'])),newline2br(option['used']))
  else:
    return "<tr><td>%s</td><td>%s</td><td>%s</td></tr>" %(option['name'],option['type'],newline2br(htmlescape(option['description'])))

def statsashtml(stat,used=True):
  if used:
    return "<tr><td>%s</td><td>%s</td></tr>" %(stat['name'],stat['used'])
  else:
    return "<tr><td>%s</td></tr>" % stat['name']

def monitorsashtml(monitor,used=True):
  if used:
    return "<tr><td>%s</td><td>%s</td></tr>" %(monitor['name'],monitor['used'])
  else:
    return "<tr><td>%s</td></tr>" % monitor['name']


def update_no_overwrite(orig,new):
  for (k,v) in new.items():
    if not(k in orig):
      orig[k] = v
        
def update_overwrite(orig,new):
  import copy
  for (k,v) in new.items():
    if not(k in orig):
      orig[k] = copy.copy(v)
    else:
      if "inherit" in v and v["inherit"]:
        orig[k]["description"] = orig[k]["description"] + "\n" +  v["description"]
        orig[k]["used"] = orig[k]["used"] + "\n" + v["used"]
      else:
        orig[k] = copy.copy(v)
        
f = open(out+'b0_options.hpp','w')

#raise Exception("")
fdiagram = open(out+'e0_diagram.hpp','w')

filemap = {}
for name,meta in sorted(metadata.items()):
  if "casadi::PluginInterface" in meta["hierarchy"] and 'casadi::FunctionInternal' not in meta["parents"]: 
    m = re.search("'(\w+)' plugin for (\w+)",meta["brief"])
    if m and "file" in meta:
      filemap["plugin_%s_%s" % ( m.group(2),m.group(1))] = (meta["file"],name)

import pickle
pickle.dump(filemap,open(out+"filemap.pkl","wb"))


# Print out doxygen information - options
for name,meta in sorted(metadata.items()):
  if not('options' in meta):
    meta['options'] = {}

  # Recursive retrieve all options from optionproviders
  optionproviders_todo = [name]
  optionproviders = []

  while optionproviders_todo:
      e = optionproviders_todo.pop(0)
      optionproviders.append(e)

      if e in metadata:
          additions = [x for x in metadata[e]["optionproviders"] if x not in optionproviders and x not in optionproviders_todo]
          optionproviders_todo.extend(additions)
  
  optionset = []
  for a in optionproviders:
    if a in metadata and 'options' in metadata[a]:
      optionset.append(metadata[a]['options'])
  
  alloptions = {}
  
  for optionprovider in reversed(optionset):
      update_overwrite(alloptions,optionprovider)
      
  myoptionskeys = alloptions.keys()
  if len(myoptionskeys)==0:
    continue
  myoptionskeys = sorted(myoptionskeys)
  
  for t in [name]:
    f.write("/// \cond INTERNAL\n")
    f.write("/** \class %s\n\\n\n\\par\n" % t)
    f.write("<a name='options'></a><table>\n")
    f.write("<caption>List of available options</caption>\n")
    f.write("<tr><th>Id</th><th>Type</th><th>Description</th><th>Used in</th></tr>\n")
    for k in myoptionskeys :
      f.write(optionsashtml(alloptions[k])+"\n")
    f.write( "</table>\n")
    f.write( "*/\n")
    f.write("/// \endcond\n")
    
    fdiagram.write("/// \cond INTERNAL\n")
    fdiagram.write("/** \class %s\n" % t)
    fdiagram.write("""<a name="diagram" id="diagram"></a><h2>Diagrams</h2>""")
    fdiagram.write( "*/\n")
    fdiagram.write("/// \endcond\n")
    

  if "casadi::PluginInterface" in meta["hierarchy"]:
    m = re.search("'(\w+)' plugin for (\w+)",meta["brief"])
    if not m:
      print("This plugin is undocumented. add \\pluginbrief{class,name} to it: " + meta["file"])
      #print meta["file"]
    else:
      f.write("/** \\addtogroup plugin_%s_%s\n\\n\n\\par\n" % (m.group(2),m.group(1)))
      f.write("<a name='options'></a><table>\n")
      f.write("<caption>List of available options</caption>\n")
      f.write("<tr><th>Id</th><th>Type</th><th>Description</th></tr>\n")
      for k in myoptionskeys :
        try:
          #if alloptions[k]["used"] in metadata and "casadi::PluginInterface" in metadata[alloptions[k]["used"]]["parents"]:
          if alloptions[k]["used"] in metadata and "casadi::PluginInterface" in metadata[metadata[alloptions[k]["used"]]["parents"][0]]["hierarchy"]:
            f.write(optionsashtml(alloptions[k],used=False)+"\n")
        except:
          pass
      f.write( "</table>\n")
      f.write( "*/\n")
    
    t = name
    f.write("/** \\addtogroup general_%s\n\\n\n\\par\n" % t.replace("casadi::","") )
    f.write("<a name='options'></a><table>\n")
    f.write("<caption>List of available options</caption>\n")
    f.write("<tr><th>Id</th><th>Type</th><th>Description</th><th>Used in</th></tr>\n")
    for k in myoptionskeys :
      f.write(optionsashtml(alloptions[k])+"\n")
    f.write( "</table>\n")
    f.write( "*/\n")
    
    fdiagram.write("/** \class %s\n" % t)
    fdiagram.write("""<a name="diagram" id="diagram"></a><h2>Diagrams</h2>""")
    fdiagram.write( "*/\n")
  if "FunctionInternal" in name:
    f.write("/** \class casadi::Function\n")
    f.write("<a name='options'></a><table>\n")
    f.write("<caption>List of available options</caption>\n")
    f.write("<tr><th>Id</th><th>Type</th><th>Description</th><th>Used in</th></tr>\n")
    for k in myoptionskeys :
      f.write(optionsashtml(alloptions[k])+"\n")
    f.write( "</table>\n")
    f.write( "*/\n")

f.close()

"""
f = file(out+'d0_stats.hpp','w')

# Print out doxygen information - stats
for name,meta in sorted(metadata.items()):
  if not('stats' in meta):
    continue

  allstats = meta['stats']
  for a in meta['hierarchy']:
    if 'stats' in metadata[a]:
      update_no_overwrite(allstats,metadata[a]['stats'])
  
  mystatskeys = allstats.keys()
  mystatskeys.sort()
  if len(mystatskeys)==0:
    continue
    
  for t in [name]:
    f.write("/// \cond INTERNAL\n")
    f.write("/** \class %s\n\\n\n\\par\n" % t)
    f.write("<a name='stats'></a><table>\n")
    f.write("<caption>List of available stats</caption>\n")
    f.write("<tr><th>Id</th><th>Used in</th></tr>\n")
    for k in mystatskeys :
      f.write(statsashtml(allstats[k])+"\n")
    f.write( "</table>\n")
    f.write( "*/\n")
    f.write("/// \endcond\n")

  if "casadi::PluginInterface" in meta["hierarchy"] and 'casadi::FunctionInternal' not in meta["parents"]:
        
    m = re.search("'(\w+)' plugin for (\w+)",meta["brief"])
    if not m:
      print "This plugin is undocumented. add \\pluginbrief{class,name} to it: " + meta["file"]
      #print meta["file"]
    else:
      f.write("/** \\addtogroup plugin_%s_%s\n\\n\n\\par\n" % (m.group(2),m.group(1)))
      f.write("<a name='stats'></a><table>\n")
      f.write("<caption>List of available stats</caption>\n")
      f.write("<tr><th>Id</th></tr>\n")
      for k in mystatskeys :
        if allstats[k]["used"] == name:
          f.write(statsashtml(allstats[k],used=False)+"\n")
      f.write( "</table>\n")
      f.write( "*/\n")
      
  if 'InternalFor' in meta:
    for t in meta['InternalFor']:
      if "casadi::PluginInterface" in meta["parents"]:
        f.write("/** \\addtogroup general_%s\n\\n\n\\par\n" % t.replace("casadi::","") )
      else:
        f.write("/** \class %s\n\\n\n\\par\n" % t)
      f.write("<a name='stats'></a><table>\n")
      f.write("<caption>List of available stats</caption>\n")
      f.write("<tr><th>Id</th><th>Used in</th></tr>\n")
      for k in mystatskeys :
        f.write(statsashtml(allstats[k])+"\n")
      f.write( "</table>\n")
      f.write( "*/\n")
    
f.close()
"""

"""
f = file(out+'c0_monitors.hpp','w')

# Print out doxygen information - monitors
for name,meta in sorted(metadata.items()):
  if not('monitors' in meta):
    continue

  allmonitors = meta['monitors']
  for a in meta['hierarchy']:
    if 'monitors' in metadata[a]:
      update_no_overwrite(allmonitors,metadata[a]['monitors'])
  
  mymonitorskeys = allmonitors.keys()
  mymonitorskeys.sort()
  
  if len(mymonitorskeys)==0:
    continue
    
  for t in [name]:
    f.write("/// \cond INTERNAL\n")
    f.write("/** \class %s\n\\n\n\\par\n" % t)
    f.write("<a name='monitors'></a><table>\n")
    f.write("<caption>List of available monitors</caption>\n")
    f.write("<tr><th>Id</th><th>Used in</th></tr>\n")
    for k in mymonitorskeys :
      f.write(monitorsashtml(allmonitors[k])+"\n")
    f.write( "</table>\n")
    f.write( "*/\n")
    f.write("/// \endcond\n")

  if "casadi::PluginInterface" in meta["hierarchy"] and 'casadi::FunctionInternal' not in meta["parents"]:
        
    m = re.search("'(\w+)' plugin for (\w+)",meta["brief"])
    if not m:
      print "This plugin is undocumented. add \\pluginbrief{class,name} to it: " + meta["file"]
      #print meta["file"]
    else:
      f.write("/** \\addtogroup plugin_%s_%s\n\\n\n\\par\n" % (m.group(2),m.group(1)))
      f.write("<a name='monitors'></a><table>\n")
      f.write("<caption>List of available monitors</caption>\n")
      f.write("<tr><th>Id</th></tr>\n")
      for k in mymonitorskeys :
        if allmonitors[k]["used"] == name:
          f.write(monitorsashtml(allmonitors[k],used=False)+"\n")
      f.write( "</table>\n")
      f.write( "*/\n")
      
  if 'InternalFor' in meta: 
    for t in meta['InternalFor']:
      if "casadi::PluginInterface" in meta["parents"]:
        f.write("/** \\addtogroup general_%s\n\\n\n\\par\n" % t.replace("casadi::","") )
      else:
        f.write("/** \class %s\n\\n\n\\par\n" % t)
      f.write("<a name='monitors'></a><table>\n")
      f.write("<caption>List of available monitors</caption>\n")
      f.write("<tr><th>Id</th><th>Used in</th></tr>\n")
      for k in mymonitorskeys :
        f.write(monitorsashtml(allmonitors[k])+"\n")
      f.write( "</table>\n")
      f.write( "*/\n")

f.close()
"""


f = open(out+'a0_schemes.hpp','w')

def enumsashtml(n,title):
  s=""
  if (n in enums):
    s+= "<a name='schemes'></a><table>\n"
    
    num = ""
    
    for i in range(len(enums[n])-1):
      m = enums[n][1+i]
      if re.search(r'_NUM_IN$',m['name']):
        num = m['name']
      if re.search(r'_NUM_OUT$',m['name']):
        num = m['name']
    s+= "<caption>%s  (%s = %d)</caption>\n" % (title,num,len(enums[n])-2)
    s+= "<tr><th>Full name</th><th>Short</th><th>Description</th></tr>\n"
    for i in range(len(enums[n])-1):
      m = enums[n][1+i]
      if re.search(r'_NUM_IN$',m['name']):
        continue
      if re.search(r'_NUM_OUT$',m['name']):
        continue
      if m['description']=='': continue
      shorthand = "_".join(m['name'].split("_")[1:]).lower()
      s+="<tr><td>%s</td><td>%s</td><td>%s</td></tr>\n" % (m['name'],shorthand,m['description'])
    s+="</table>\n"
  return s

# Write out all input/output information
for scheme in enums:
  types = ["input","output","struct"]
  for t in types:
    if re.search(t,scheme,re.IGNORECASE):
      f.write("/** \defgroup scheme_%s\n" % scheme)
      f.write(enumsashtml(scheme,"%s scheme: casadi::%s" % (t.capitalize(),scheme)))
      f.write( "*/\n")

# Print out doxygen information - schemes
for name,meta in sorted(metadata.items()):

  inputscheme = None
  outputscheme = None
  if not('inputscheme' in meta):
    for a in meta['hierarchy']:
      if a in metadata and 'inputscheme' in metadata[a]:
        inputscheme = metadata[a]['inputscheme']
  else:
    inputscheme = meta['inputscheme']

  if not('outputscheme' in meta):
    for a in meta['hierarchy']:
      if  a in metadata and 'outputscheme' in metadata[a]:
        outputscheme = metadata[a]['outputscheme']
  else:
    outputscheme = meta['outputscheme']
  
  targets = [name]
  if 'InternalFor' in meta:
    targets+=meta['InternalFor']
    
  for t in [name]:
    if not(inputscheme is None) or not(outputscheme is None):
      f.write("/// \cond INTERNAL\n")
      f.write("/** \class %s\n\\n\n\\par\n" % t)
      if not(inputscheme is None) and not(outputscheme is None):
        f.write("@copydoc scheme_%s\n" % inputscheme)
        f.write("<br/>\n")
        f.write("@copydoc scheme_%s\n" % outputscheme)
      elif outputscheme is None:
        f.write("@copydoc scheme_%s\n" % inputscheme)
      elif inputscheme is None:
        f.write("@copydoc scheme_%s\n" % outputscheme)
      f.write( "*/\n")
      f.write("/// \endcond\n")
      
  if 'InternalFor' in meta:
    for t in meta['InternalFor']:
      if not(inputscheme is None) or not(outputscheme is None):
        if "casadi::PluginInterface" in meta["parents"]:
          f.write("/** \\addtogroup general_%s\n\\n\n\\par\n" % t.replace("casadi::","") )
        else:
          f.write("/** \class %s\n\\n\n\\par\n" % t)
        if not(inputscheme is None) and not(outputscheme is None):
          f.write("@copydoc scheme_%s\n" % inputscheme)
          f.write("<br/>\n")
          f.write("@copydoc scheme_%s\n" % outputscheme)
        elif outputscheme is None:
          f.write("@copydoc scheme_%s\n" % inputscheme)
        elif inputscheme is None:
          f.write("@copydoc scheme_%s\n" % outputscheme)
        f.write( "*/\n")
   
f.close()



#
#     MIT No Attribution
#
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl, KU Leuven.
#
#     Permission is hereby granted, free of charge, to any person obtaining a copy of this
#     software and associated documentation files (the "Software"), to deal in the Software
#     without restriction, including without limitation the rights to use, copy, modify,
#     merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
#     permit persons to whom the Software is furnished to do so.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
#     INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
#     PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#     OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from pylab import *

# End time
T = 10.

# Number of control intervals
N = 20

# Number of Runge-Kutta 4 steps per interval and step size
NK = 20
DT = T/(N*NK)

# Number of discrete control values
NU = 101

# Number of discrete state values
NX = 101

# System dynamics, can be called with matricex
def f(x1,x2,u):
  x1_dot = (1 - x2*x2)*x1 - x2 + u
  x2_dot = x1
  q_dot  = x1*x1 + x2*x2 + u*u
  return (x1_dot, x2_dot, q_dot)

# Control enumeration
U  = linspace(-1,1,NU)

# State space enumeration
x1 = linspace(-1,1,NX)
x2 = linspace(-1,1,NX)
X1,X2 = meshgrid(x1,x2)

# For each control action and state, precalculate next state and stage cost
stage_J = []
next_x1 = []
next_x2 = []
for u in U:
  # Take number of integration steps
  X1_k = copy(X1)
  X2_k = copy(X2)
  Q_k = zeros(X1.shape)
  for k in range(NK):
    # RK4 integration for x1, x2 and q
    k1_x1, k1_x2, k1_q = f(X1_k,                X2_k,                u)
    k2_x1, k2_x2, k2_q = f(X1_k + DT/2 * k1_x1, X2_k + DT/2 * k1_x2, u)
    k3_x1, k3_x2, k3_q = f(X1_k + DT/2 * k2_x1, X2_k + DT/2 * k2_x2, u)
    k4_x1, k4_x2, k4_q = f(X1_k + DT   * k3_x1, X2_k + DT   * k3_x2, u)
    X1_k += DT/6*(k1_x1 + 2*k2_x1 + 2*k3_x1 + k4_x1)
    X2_k += DT/6*(k1_x2 + 2*k2_x2 + 2*k3_x2 + k4_x2)
    Q_k  += DT/6*(k1_q  + 2*k2_q  + 2*k3_q  + k4_q )

  # Find out which state comes next (index)
  X1_k = matrix.round((X1_k+1)/2*(NX-1)).astype(int)
  X2_k = matrix.round((X2_k+1)/2*(NX-1)).astype(int)

  # Infinite cost if state gets out-of-bounds
  I = X1_k  <  0; Q_k[I]=inf; X1_k[I]=0
  I = X2_k  <  0; Q_k[I]=inf; X2_k[I]=0
  I = X1_k >= NX; Q_k[I]=inf; X1_k[I]=0
  I = X2_k >= NX; Q_k[I]=inf; X2_k[I]=0

  # Save the stage cost and next state
  next_x1.append(X1_k)
  next_x2.append(X2_k)
  stage_J.append(Q_k)

# Calculate cost-to-go (no end cost) and optimal control
J = zeros(X1.shape)
U_opt = []
for k in reversed(list(range(N))):
  # Cost to go for the previous step, optimal control action
  J_prev = inf*ones(X1.shape)
  u_prev = -ones(X1.shape,dtype=int)

  # Test all control actions
  for uind in range(NU):
    J_prev_test = J[next_x2[uind],next_x1[uind]]+stage_J[uind]
    better = J_prev_test<J_prev
    u_prev[better] = uind
    J_prev[better] = J_prev_test[better]

  # Update cost-to-go and save optimal control
  J = J_prev
  U_opt.append(u_prev)

# Reorder U_opt by stage
U_opt.reverse()

# Find optimal control starting at x1=0, x2=1
i1 = NX//2
i2 = NX-1
u_opt = []
x1_opt = [x1[i1]]
x2_opt = [x2[i2]]
cost = 0
for k in range(N):
  # Get the optimal control and go to next step
  u_ind = U_opt[k][i2,i1]
  cost += stage_J[u_ind][i2,i1]
  i1, i2 = next_x1[u_ind][i2,i1], next_x2[u_ind][i2,i1]

  # Save the trajectories
  u_opt.append(U[u_ind])
  x1_opt.append(x1[i1])
  x2_opt.append(x2[i2])

# Optimal cost
print("Minimal cost: ", cost)
assert abs(cost-J[NX-1,NX//2])<1e-8 # Consistency check

# Plot
figure(1)
clf()

# Plot optimal cost-to-go
subplot(121)
contourf(X1,X2,J)
colorbar()
xlabel('x1')
ylabel('x2')
title('Cost-to-go')

subplot(122)
plot(linspace(0,T,N+1),x1_opt,'--')
plot(linspace(0,T,N+1),x2_opt,'-.')
step(linspace(0,T,N),u_opt,'-')
plt.title("Dynamic programming solution")
plt.xlabel('time')
plt.legend(['x1 trajectory','x2 trajectory','u trajectory'])
grid(True)
show()



#
#     MIT No Attribution
#
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl, KU Leuven.
#
#     Permission is hereby granted, free of charge, to any person obtaining a copy of this
#     software and associated documentation files (the "Software"), to deal in the Software
#     without restriction, including without limitation the rights to use, copy, modify,
#     merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
#     permit persons to whom the Software is furnished to do so.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
#     INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
#     PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#     OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
from casadi import *

# In this example, we fit a nonlinear model to measurements
#
# This example uses more advanced constructs than the vdp* examples:
# Since the number of control intervals is potentially very large here,
# we use memory-efficient Map and MapAccum, in combination with
# codegeneration.
#
# We will be working with a 2-norm objective:
# || y_measured - y_simulated ||_2^2
#
# This form is well-suited for the Gauss-Newton Hessian approximation.

############ SETTINGS #####################
N = 10000  # Number of samples
fs = 610.1 # Sampling frequency [hz]

param_truth = DM([5.625e-6,2.3e-4,1,4.69])
param_guess = DM([5,2,1,5])
scale = vertcat(1e-6,1e-4,1,1)

############ MODELING #####################
y  = MX.sym('y')
dy = MX.sym('dy')
u  = MX.sym('u')

states = vertcat(y,dy);
controls = u;

M = MX.sym("x")
c = MX.sym("c")
k = MX.sym("k")
k_NL = MX.sym("k_NL")

params = vertcat(M,c,k,k_NL)

rhs = vertcat(dy , (u-k_NL*y**3-k*y-c*dy)/M)

# Form an ode function
ode = Function('ode',[states,controls,params],[rhs])

############ Creating a simulator ##########
N_steps_per_sample = 10
dt = 1/fs/N_steps_per_sample

# Build an integrator for this system: Runge Kutta 4 integrator
k1 = ode(states,controls,params)
k2 = ode(states+dt/2.0*k1,controls,params)
k3 = ode(states+dt/2.0*k2,controls,params)
k4 = ode(states+dt*k3,controls,params)

states_final = states+dt/6.0*(k1+2*k2+2*k3+k4)

# Create a function that simulates one step propagation in a sample
one_step = Function('one_step',[states, controls, params],[states_final]);

X = states;

for i in range(N_steps_per_sample):
  X = one_step(X, controls, params)

# Create a function that simulates all step propagation on a sample
one_sample = Function('one_sample',[states, controls, params], [X])

############ Simulating the system ##########
all_samples = one_sample.mapaccum("all_samples", N)

# Choose an excitation signal
numpy.random.seed(0)
u_data = DM(0.1*numpy.random.random(N))

x0 = DM([0,0])
X_measured = all_samples(x0, u_data, repmat(param_truth,1,N))

y_data = X_measured[0,:].T

# You may add some noise here
#y_data+= 0.001*numpy.random.random(N)
# When noise is absent, the fit will be perfect.

# Use just-in-time compilation to speed up the evaluation
if Importer.has_plugin('clang'):
  with_jit = True
  compiler = 'clang'
elif Importer.has_plugin('shell'):
  with_jit = True
  compiler = 'shell'
else:
  print("WARNING; running without jit. This may result in very slow evaluation times")
  with_jit = False
  compiler = ''

############ Create a Gauss-Newton solver ##########
def gauss_newton(e,nlp,V):
  J = jacobian(e,V)
  H = triu(mtimes(J.T, J))
  sigma = MX.sym("sigma")
  hessLag = Function('nlp_hess_l',{'x':V,'lam_f':sigma, 'hess_gamma_x_x':sigma*H},
                     ['x','p','lam_f','lam_g'], ['hess_gamma_x_x'],
                     dict(jit=with_jit, compiler=compiler))
  return nlpsol("solver","ipopt", nlp, dict(hess_lag=hessLag, jit=with_jit, compiler=compiler))

############ Identifying the simulated system: single shooting strategy ##########

# Note, it is in general a good idea to scale your decision variables such
# that they are in the order of ~0.1..100
X_symbolic = all_samples(x0, u_data, repmat(params*scale,1,N))

e = y_data-X_symbolic[0,:].T;
nlp = {'x':params, 'f':0.5*dot(e,e)}

solver = gauss_newton(e,nlp, params)

sol = solver(x0=param_guess)

print(sol["x"]*scale)

assert(norm_inf(sol["x"]*scale-param_truth)<1e-8)

############ Identifying the simulated system: multiple shooting strategy ##########

# All states become decision variables
X = MX.sym("X", 2, N)

Xn = one_sample.map(N, 'openmp')(X, u_data.T, repmat(params*scale,1,N))

gaps = Xn[:,:-1]-X[:,1:]

e = y_data-Xn[0,:].T;

V = veccat(params, X)

nlp = {'x':V, 'f':0.5*dot(e,e),'g': vec(gaps)}

# Multipleshooting allows for careful initialization
yd = np.diff(y_data,axis=0)*fs
X_guess = horzcat(y_data , vertcat(yd,yd[-1])).T;

x0 = veccat(param_guess,X_guess)

solver = gauss_newton(e,nlp, V)

sol = solver(x0=x0,lbg=0,ubg=0)

print(sol["x"][:4]*scale)

assert(norm_inf(sol["x"][:4]*scale-param_truth)<1e-8)



#     This file is part of CasADi.
#
#     CasADi -- A symbolic framework for dynamic optimization.
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl,
#                             KU Leuven. All rights reserved.
#     Copyright (C) 2011-2014 Greg Horn
#
#     CasADi is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     CasADi is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

# Linear solvers
# =================
#
# We demonstrate solving a dense system A.x=b by using different linear solvers.

from casadi import *
from numpy import *
import time

n=100

# We generate $A \in \mathbf{R}^{n \times n}$, $x \in \mathbf{R}^{n}$ with $n=100$

A=DM([[cos(i*j)-sin(i) for i in range(n)] for j in range(n)])
x=DM([tan(i) for i in range(n)])

# We generate the b vector:

b= mtimes(A,x)

# Commented out pendling completion #1615

# # We demonstrate the LinearSolver API with CSparse:
# s = LinearSolver("s", "csparse", A.sparsity())

# # Give it the matrix A
# s.setInput(A,"A")
# # Do the LU factorization
# s.prepare()

# # Give it the matrix b
# s.setInput(b,"B")

# # And we are off to find x...
# s.solve()

# x_ = s.getOutput("X")

# # By looking at the residuals between the x we knew in advance and the computed x, we see that the CSparse solver works
# print "Sum of residuals = %.2e" % norm_1(x-x_)

# # Comparison of different linear solvers
# # ======================================
# for solver in ("lapacklu","lapackqr","csparse"):
#   s = LinearSolver("s", solver, A.sparsity()) # We create a solver

#   s.setInput(A,"A") # Give it the matrix A
  
#   t0 = time.time()
#   for i in range(100):
#     s.prepare()        # Do the LU factorization
#   pt = (time.time()-t0)/100

#   s.setInput(b,"B")  # Give it the matrix b

#   t0 = time.time()
#   for i in range(100):
#     s.solve()
#   st = (time.time()-t0)/100
  
#   x_ = s.getOutput("X")

#   print ""
#   print solver
#   print "=" * 10
#   print "Sum of residuals = %.2e" % norm_1(x-x_)
#   print "Preparation time = %0.2f ms" % (pt*1000)
#   print "Solve time       = %0.2f ms" % (st*1000)
#   assert(norm_1(x-x_)<1e-9)
  
# # Note that these 



#
#     MIT No Attribution
#
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl, KU Leuven.
#
#     Permission is hereby granted, free of charge, to any person obtaining a copy of this
#     software and associated documentation files (the "Software"), to deal in the Software
#     without restriction, including without limitation the rights to use, copy, modify,
#     merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
#     permit persons to whom the Software is furnished to do so.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
#     INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
#     PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#     OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# -*- coding: utf-8 -*-
from casadi import *

# minimize    3x + 4y
# subject to  x + 2y <= 14
#            3x -  y >= 0
#             x -  y <= 2


# Sparsity of the LP linear term
A = Sparsity.dense(3, 2)

# Create solver
solver = conic('solver', 'qpoases', {'a':A})
#solver = conic('solver', 'clp', {'a':A}) # Use clp

g = DM([3,4])
a = DM([[1, 2],[3, -1], [1, -1]])
lba = DM([-inf, 0, -inf])
uba = DM([14, inf, 2])

sol = solver(g=g, a=a, lba=lba, uba=uba)
print(sol)



#     This file is part of CasADi.
#
#     CasADi -- A symbolic framework for dynamic optimization.
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl,
#                             KU Leuven. All rights reserved.
#     Copyright (C) 2011-2014 Greg Horn
#
#     CasADi is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     CasADi is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

from casadi import *

# A simple case of Callback
# ================================

# Callback allows the user to create functions that can be embedded into
# CasADi expressions. The user creates a class that inherits from this class
# and implements a subset of the virtual methods.
# Although Callback itself is implemented is C++, the virtual methods can be
# implemented in Python or MATLAB thanks to cross-language polymorphism as
# supported by the SWIG framework.

class Fac(Callback):
  def __init__(self, name, opts={}):
    Callback.__init__(self)
    self.construct(name, opts)

  def get_n_in(self): return 1
  def get_n_out(self): return 1

  def eval(self, arg):
    x = arg[0]
    y = 1
    for i in range(int(x)):
      y*=(i+1)
    return [y]

fac = Fac('fac')

# Evaluate numerically
y = fac(4)

print("4! = ", y)

# Using the function in a graph
# ==============================

x = MX.sym("x")
y = fac(x)

f = Function('f', [x],[y])

y = f(5)

print("5! = ", y)



#
#     MIT No Attribution
#
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl, KU Leuven.
#
#     Permission is hereby granted, free of charge, to any person obtaining a copy of this
#     software and associated documentation files (the "Software"), to deal in the Software
#     without restriction, including without limitation the rights to use, copy, modify,
#     merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
#     permit persons to whom the Software is furnished to do so.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
#     INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
#     PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#     OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
from casadi import *

"""
This example mainly intended for CasADi presentations. It contains a compact
implementation of a direct single shooting method for DAEs using a minimal
number of CasADi concepts.

It solves the following optimal control problem (OCP) in differential-algebraic
equations (DAE):

minimize     integral_{t=0}^{10} x0^2 + x1^2 + u^2  dt
x0,x1,z,u

subject to   dot(x0) == z*x0-x1+u     \
             dot(x1) == x0             }  for 0 <= t <= 10
                   0 == x1^2 + z - 1  /
             x0(t=0) == 0
             x1(t=0) == 1
             x0(t=10) == 0
             x1(t=10) == 0
             -0.75 <= u <= 1  for 0 <= t <= 10

Note that other methods such as direct collocation or direct multiple shooting
are usually preferably to the direct single shooting method in practise.

Joel Andersson, 2012-2015
"""

# Declare variables
x = SX.sym("x",2) # Differential states
z = SX.sym("z")   # Algebraic variable
u = SX.sym("u")   # Control

# Differential equation
f_x = vertcat(z*x[0]-x[1]+u, x[0])

# Algebraic equation
f_z = x[1]**2 + z - 1

# Lagrange cost term (quadrature)
f_q = x[0]**2 + x[1]**2 + u**2

# Create an integrator
dae = {'x':x, 'z':z, 'p':u, 'ode':f_x, 'alg':f_z, 'quad':f_q}
# interval length 0.5s
I = integrator('I', "idas", dae, 0, 0.5)

# All controls
U = MX.sym("U", 20)

# Construct graph of integrator calls
X  = [0,1]
J = 0
for k in range(20):
  Ik = I(x0=X, p=U[k])
  X = Ik['xf']
  J += Ik['qf']   # Sum up quadratures

# Allocate an NLP solver
nlp = {'x':U, 'f':J, 'g':X}
opts = {"ipopt.linear_solver":"ma27"}
solver = nlpsol("solver", "ipopt", nlp, opts)

# Pass bounds, initial guess and solve NLP
sol = solver(lbx = -0.75, # Lower variable bound
             ubx =  1.0,  # Upper variable bound
             lbg =  0.0,  # Lower constraint bound
             ubg =  0.0,  # Upper constraint bound
             x0  =  0.0) # Initial guess
print(sol)



#
#     MIT No Attribution
#
#     Copyright (C) 2010-2023 Joel Andersson, Joris Gillis, Moritz Diehl, KU Leuven.
#
#     Permission is hereby granted, free of charge, to any person obtaining a copy of this
#     software and associated documentation files (the "Software"), to deal in the Software
#     without restriction, including without limitation the rights to use, copy, modify,
#     merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
#     permit persons to whom the Software is furnished to do so.
#
#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
#     INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
#     PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#     HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#     OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#     SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
from casadi import *
import numpy as np
import matplotlib.pyplot as plt

# Consider the following nonlinear system with two states and one control input:
x1 = MX.sym('x1')
x2 = MX.sym('x2')
x = vertcat(x1, x2)
u = MX.sym('u')
xdot = vertcat((1-x2**2)*x1 - x2 + u, x1)

# In addition to the dynamics, we are also interested in the sum-of-squares distance from the origin:
y = x1**2 + x2**2

# We want to simulate this system over T seconds, with the control discretized into N intervals:
# Time horizon, discretization
N = 10
T = 10.
tgrid = np.linspace(0, T, N+1)

# We will integrate the problem using CVODES, while also calculating the integral of y
dae = {'x':x, 'u':u, 'ode':xdot, 'quad':y}
F = integrator('F', 'cvodes', dae, tgrid[0], tgrid[1:])

# We are also interested in the pointwise values of y, so let's create a function for that too
yfun = Function('yfun', [x], [y], ['x'], ['y'])

# Let's define initial conditions for x and values for the piecewise constant controls
u = np.linspace(-1, 1, N)
x0 = np.array([0, 0])

# We can simulate the system as follows:
Fk = F(x0 = x0, u = u)
xf = Fk['xf']
qf = Fk['qf']

# Let's add the state at the initial time to xf and qf
xf = horzcat(x0, xf)
qf = horzcat(0, qf)

# Call yfun to get y at each grid point
yf = yfun(xf)

# Convert CasADi matrices to numpy arrays
x_sim = xf.full()
y_sim = yf.full()
q_sim = qf.full()

# Plot simulation trajectories and pointwise sum of squares contribution to y, which we compare with the integral of y
fig, [ax1, ax2] = plt.subplots(nrows = 1, ncols = 2)

ax1.plot(tgrid, x_sim[0,:], '*-', label = '$x_1(t)$')
ax1.plot(tgrid, x_sim[1,:], '*-', label = '$x_2(t)$')
ax1.plot(tgrid, y_sim[0,:], '*-', label = '$(x_1(t))^2 + (x_2(t))^2$')
ax1.step(tgrid, np.insert(u, 0, np.nan), label = '$u(t)$')
ax1.set_xlabel(r'$t$')
ax1.set_title('Trajectories')
ax1.legend()
ax1.grid()

ax2.plot(tgrid, q_sim[0,:], '*-', label = r'$\int_{t=0}^{t_k}{((x_1(t))^2 + (x_2(t))^2) \, dt}$')
ax2.plot(tgrid, T/N * np.cumsum(y_sim), '*-', label = r'$\sum_{i=0}^{k}{((x_1(t_i))^2 + (x_2(t_i))^2) \, \frac{T}{N}}$')
ax2.set_xlabel(r'$t_k$')
ax2.set_title('Integral cost')
ax2.legend()
ax2.grid()



plt.show()
